# python 2/3
from __future__ import division, print_function, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa

from builtins import range
from pyrocko import trace, util, model, pile
import unittest
import math
import time
import numpy as num
import pickle as pickle

from . import common

sometime = 1234567890.
d2r = num.pi/180.


def numeq(a, b, eps):
    return num.all(num.abs(num.array(a) - num.array(b)) <= eps)


def floats(l):
    return num.array(l, dtype=num.float)


def is_multiple(x, d, eps):
    x2 = int(round(x / d)) * d
    return abs(x - x2) < d*eps


class TraceTestCase(unittest.TestCase):

    def testIntegrationDifferentiation(self):

        fmin = 0.1
        fmax = 2.0
        for tlen in [100., 1000., 10000]:
            for dt in [0.1, 0.05]:
                for f in [0.2, 0.5, 1.0]:

                    n = int(tlen/dt)
                    tfade = 2.0/fmin

                    xdata = num.arange(n)*dt
                    ydata = num.sin(xdata*2.*num.pi*f)
                    a = trace.Trace(channel='A', deltat=dt, ydata=ydata)

                    b = a.transfer(
                        tfade, (fmin/2., fmin, fmax, fmax*2.),
                        transfer_function=trace.IntegrationResponse())
                    b.set_codes(channel='B')

                    c = a.transfer(
                        tfade, (fmin/2., fmin, fmax, fmax*2.),
                        transfer_function=trace.DifferentiationResponse())
                    c.set_codes(channel='C')

                    eps = 0.005
                    xdata = b.get_xdata()
                    ydata = b.get_ydata()
                    ydata_shouldbe = -num.cos(xdata*2*num.pi*f) / (2.*num.pi*f)

                    # if num.amax(num.abs(ydata-ydata_shouldbe)) > eps:
                    # b_shouldbe = trace.Trace(
                    #     channel='B', location='integrated',
                    #     deltat=dt, ydata=ydata_shouldbe, tmin=xdata[0])
                    #     trace.snuffle([b, b_shouldbe])

                    assert num.amax(num.abs(ydata-ydata_shouldbe)) < eps, \
                        'integration failed'

                    xdata = c.get_xdata()
                    ydata = c.get_ydata()
                    ydata_shouldbe = num.cos(xdata*2*num.pi*f) * (2.*num.pi*f)

                    # if num.amax(num.abs(ydata-ydata_shouldbe)) > eps:
                    # c_shouldbe = trace.Trace(
                    #     channel='C', location='differentiated',
                    #     deltat=dt, ydata=ydata_shouldbe, tmin=xdata[0])
                    #     trace.snuffle([c, c_shouldbe])

                    assert num.amax(num.abs(ydata-ydata_shouldbe)) < eps, \
                        'differentiation failed'

    def testDegapping(self):
        dt = 1.0
        atmin = 100.

        for btmin in range(90, 120):
            a = trace.Trace(deltat=dt, ydata=num.zeros(10), tmin=atmin)
            b = trace.Trace(deltat=dt, ydata=num.ones(5), tmin=btmin)
            traces = [a, b]
            traces.sort(key=lambda a: a.full_id)
            xs = trace.degapper(traces)

            if btmin == 90:
                assert len(xs) == 2
            elif btmin > 90 and btmin < 115:
                assert len(xs) == 1
            else:
                assert len(xs) == 2

        a = trace.Trace(deltat=dt, ydata=num.zeros(10), tmin=100)
        b = trace.Trace(deltat=dt, ydata=num.ones(10), tmin=100)
        traces = [a, b]
        traces.sort(key=lambda a: a.full_id)
        xs = trace.degapper(traces)
        assert len(xs) == 1
        for x in xs:
            assert x.tmin == 100
            assert x.get_ydata().size == 10

        for meth, res in (
                ('use_second', [1., 1.]),
                ('use_first', [0., 0.]),
                ('crossfade_cos', [0.25, 0.75]),
                ('add', [0., 0.])):

            if meth == 'add':
                a = trace.Trace(deltat=dt, ydata=num.ones(10)*-1, tmin=100)
            else:
                a = trace.Trace(deltat=dt, ydata=num.zeros(10), tmin=100)
            b = trace.Trace(deltat=dt, ydata=num.ones(10), tmin=108)
            traces = [a, b]
            traces.sort(key=lambda a: a.full_id)
            xs = trace.degapper(traces, deoverlap=meth)
            for x in xs:
                assert x.ydata.size == 18
                assert numeq(x.ydata[8:10], res, 1e-6)

    def testRotation(self):
        s2 = math.sqrt(2.)
        ndata = num.array([s2, s2], dtype=num.float)
        edata = num.array([s2, 0.], dtype=num.float)
        dt = 1.0
        n = trace.Trace(deltat=dt, ydata=ndata, tmin=100, channel='N')
        e = trace.Trace(deltat=dt, ydata=edata, tmin=100, channel='E')
        rotated = trace.rotate([n, e], 45., ['N', 'E'], ['R', 'T'])
        for tr in rotated:
            if tr.channel == 'R':
                r = tr
            if tr.channel == 'T':
                t = tr

        assert numeq(r.get_ydata(), [2., 1.], 1.0e-6)
        assert numeq(t.get_ydata(), [0., -1], 1.0e-6)

    def testLQTRotation(self):
        ba = 225.
        inci = 90.-num.arctan(1./num.sqrt(2.))*180./num.pi
        in_channels = list('ZEN')
        zdata = num.ones(2)
        edata = num.ones(2)
        ndata = num.ones(2)
        z = trace.Trace(deltat=1., ydata=zdata, tmin=1., channel='Z')
        e = trace.Trace(deltat=1., ydata=edata, tmin=1., channel='E')
        n = trace.Trace(deltat=1., ydata=ndata, tmin=1., channel='N')
        l, q, t = trace.rotate_to_lqt([z, n, e], ba, inci, in_channels)
        assert(num.all(l.get_ydata() - num.sqrt(3.) < 1.0e-12))
        assert(num.all(t.get_ydata() < 1.0e-12))
        assert(num.all(q.get_ydata() < 1.0e-12))

    def testProjection(self):
        s2 = math.sqrt(2.)
        ndata = num.array([s2, s2], dtype=num.float)
        edata = num.array([s2, 0.], dtype=num.float)
        ddata = num.array([1., -1.], dtype=num.float)
        dt = 1.0
        n = trace.Trace(deltat=dt, ydata=ndata, tmin=100, channel='N')
        e = trace.Trace(deltat=dt, ydata=edata, tmin=100, channel='E')
        d = trace.Trace(deltat=dt, ydata=ddata, tmin=100, channel='D')
        azi = 45.
        cazi = math.cos(azi*d2r)
        sazi = math.sin(azi*d2r)
        rot45 = num.array(
            [[cazi, sazi, 0], [-sazi, cazi, 0], [0, 0, -1]], dtype=num.float)

        def C(x):
            return model.Channel(x)

        rotated = trace.project(
            [n, e, d],
            rot45,
            [C('N'), C('E'), C('D')],
            [C('R'), C('T'), C('U')])

        for tr in rotated:
            if tr.channel == 'R':
                r = tr
            if tr.channel == 'T':
                t = tr
            if tr.channel == 'U':
                u = tr

        assert(num.all(r.get_ydata() - num.array([2., 1.]) < 1.0e-6))
        assert(num.all(t.get_ydata() - num.array([0., -1]) < 1.0e-6))
        assert(num.all(u.get_ydata() - num.array([-1., 1.]) < 1.0e-6))

        deps = trace.project_dependencies(
            rot45,
            [C('N'), C('E'), C('D')],
            [C('R'), C('T'), C('U')])

        assert (
            set(['N', 'E']) == set(deps['R'])
            and set(['N', 'E']) == set(deps['T'])
            and set(['D']) == set(deps['U']))

        # should work though no horizontals given
        trace.project(
            [d],
            rot45,
            [C('N'), C('E'), C('D')], [C('R'), C('T'), C('U')])

        if tr.channel == 'U':
            u = tr

        assert(num.all(u.get_ydata() - num.array([-1., 1.]) < 1.0e-6))

    def testExtend(self):
        tmin = sometime
        t = trace.Trace(tmin=tmin, ydata=num.ones(10, dtype=num.float))
        tmax = t.tmax
        t.extend(tmin-10.2, tmax+10.7)
        assert int(round(tmin-t.tmin)) == 10
        assert int(round(t.tmax-tmax)) == 11
        assert num.all(t.ydata[:10] == num.zeros(10, dtype=num.float))
        assert num.all(t.ydata[-11:] == num.zeros(11, dtype=num.float))
        assert num.all(t.ydata[10:-11] == num.ones(10, dtype=num.float))

        t = trace.Trace(tmin=tmin, ydata=num.arange(10, dtype=num.float)+1.)
        t.extend(tmin-10.2, tmax+10.7, fillmethod='repeat')
        assert num.all(t.ydata[:10] == num.ones(10, dtype=num.float))
        assert num.all(t.ydata[-11:] == num.zeros(11, dtype=num.float)+10.)
        assert num.all(t.ydata[10:-11] == num.arange(10, dtype=num.float)+1.)

    def testAppend(self):
        a = trace.Trace(ydata=num.zeros(0, dtype=num.float), tmin=sometime)
        for i in range(10000):
            a.append(num.arange(1000, dtype=num.float))
        assert a.ydata.size == 10000*1000

    def testDownsampling(self):

        data = [
            (1./125., 1/10., True)]

        for f in [0.1, 1.0, 10.]:
            for i in range(1, 6):
                data.append((f*float(1.), f*float(i), False))

        for (dt1, dt2, test_fail) in data:

            for toff in [
                    num.random.random() * dt1 + num.random.random() * dt2]:

                n = 1024
                xdata = num.arange(n, dtype=num.float)
                ydata = num.exp(-((xdata-n/2)/10.)**2)

                for orig_snap in (True, False):
                    t = trace.Trace(
                        ydata=ydata,
                        tmin=sometime+toff,
                        deltat=dt1,
                        location='1')

                    if orig_snap:
                        t.snap()

                    t2 = t.copy()
                    if test_fail:
                        with self.assertRaises(trace.UnavailableDecimation):
                            t2.downsample_to(dt2, allow_upsample_max=1)

                    for snap in (True, False):
                        t2 = t.copy()
                        t2.set_codes(location='2')
                        t2.downsample_to(dt2, allow_upsample_max=10, snap=snap)

                        # print(
                        #     "%20.4f %6s %10.4f %10.4f %20.4f %20.4f %6s" % (
                        #         toff, snap,
                        #         t.deltat, t2.deltat,
                        #         t.tmin, t2.tmin,
                        #         is_multiple(t2.tmin, t2.deltat, 1e-3)))

                        if snap and orig_snap:
                            assert is_multiple(t2.tmin, t2.deltat, 1e-3)
                        elif snap and not orig_snap:
                            assert abs(
                                int(round(t2.tmin / t2.deltat)) *
                                t2.deltat - t2.tmin) <= 0.51*t.deltat

            # trace.snuffle([t, t2])

    def testFiltering(self):
        tmin = sometime
        # b = time.time()
        for i in range(100):
            n = 10000
            t = trace.Trace(
                tmin=tmin, deltat=0.05, ydata=num.ones(n, dtype=num.float))
            t.lowpass(4,  5.)
            t.highpass(4, 0.1)
        # d1 = time.time() - b
        # b = time.time()
        for i in range(100):
            n = 10000
            t = trace.Trace(
                tmin=tmin, deltat=0.05, ydata=num.ones(n, dtype=num.float))
            t.bandpass_fft(0.1, 5.)
        # d2 = time.time() - b

    def testCropping(self):
        n = 20
        tmin = sometime
        t = trace.Trace(
            tmin=tmin, deltat=0.05, ydata=num.ones(n, dtype=num.float))
        tmax = t.tmax
        t.ydata[:3] = 0.0
        t.ydata[-3:] = 0.0
        t.crop_zeros()
        assert abs(t.tmin - (tmin + 0.05*3)) < 0.05*0.01
        assert abs(t.tmax - (tmax - 0.05*3)) < 0.05*0.01
        t = trace.Trace(
            tmin=tmin, deltat=0.05, ydata=num.zeros(n, dtype=num.float))
        ok = False
        try:
            t.crop_zeros()
        except trace.NoData:
            ok = True

        assert ok

        t = trace.Trace(
            tmin=tmin, deltat=0.05, ydata=num.ones(n, dtype=num.float))
        t.crop_zeros()
        assert t.ydata.size == 20

    def testAdd(self):
        tmin = sometime
        deltat = 0.05
        for distortion in [-0.2, 0.2]:
            for ioffs, result in [(1, [0., 1., 1., 1., 0.]),
                                  (-1, [1., 1., 0., 0., 0.]),
                                  (3, [0., 0., 0., 1., 1.]),
                                  (10, [0., 0., 0., 0., 0.]),
                                  (-10, [0., 0., 0., 0., 0.])]:

                a = trace.Trace(
                    tmin=tmin, deltat=deltat,
                    ydata=num.zeros(5, dtype=num.float))
                b = trace.Trace(
                    tmin=tmin+(ioffs+distortion)*deltat, deltat=deltat,
                    ydata=num.ones(3, dtype=num.float))

                a.add(b, interpolate=False)
                assert numeq(a.ydata, result, 0.001)

    def testAdd2(self):
        for toff in num.random.random(100):
            t1 = trace.Trace(
                tmin=sometime, deltat=0.05,
                ydata=num.ones(5, dtype=num.float))
            t2 = trace.Trace(
                tmin=sometime-toff*0.1, deltat=0.05,
                ydata=num.ones(5, dtype=num.float))
            t1.add(t2, interpolate=False)

    def testPeaks(self):
        n = 1000
        t = trace.Trace(
            tmin=0, deltat=0.1, ydata=num.zeros(n, dtype=num.float))
        t.ydata[1] = 1.
        t.ydata[500] = 1.
        t.ydata[999] = 1.
        tp, ap = t.peaks(0.5, 1.)
        assert numeq(tp, [0.1, 50., 99.9], 0.0001)
        assert numeq(ap, [1., 1., 1.], 0.0001)

        t.ydata[504] = 1.0
        t.ydata[511] = 1.0

        tp, ap = t.peaks(0.5, 1.)

        assert numeq(tp, [0.1, 50, 51.1, 99.9], 0.0001)
        assert numeq(ap, [1., 1., 1., 1.], 0.0001)

    def testCorrelate(self):
        for la, lb, mode, res in [
                ([0, 1, .5, 0, 0],    [0, 0, 0, 1, 0],    'same', 0.3),
                ([0, 1, .5, 0, 0, 0], [0, 1, 0],          'valid', 0.1),
                ([0, 1, .5],          [0, 0, 0, 1, 0, 0], 'full', 0.3)]:

            for ia in range(5):
                for ib in range(5):

                    ya = floats(la + [0] * ia)
                    yb = floats(lb + [0] * ib)

                    a = trace.Trace(tmin=10., deltat=0.1, ydata=ya)
                    b = trace.Trace(tmin=10.1, deltat=0.1, ydata=yb)
                    c = trace.correlate(a, b, mode=mode)

                    assert numeq(c.max(), [res, 1.], 0.0001)

    def testCorrelateNormalization(self):

        ya = floats([1, 2, 1])
        yb = floats([0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 2, 1])

        a = trace.Trace(tmin=sometime, deltat=0.1, ydata=ya)
        b = trace.Trace(tmin=sometime, deltat=0.1, ydata=yb)

        c_ab = trace.correlate(a, b)
        c_ab2 = trace.correlate(a, b, normalization='gliding')
        c_ba = trace.correlate(b, a)
        c_ba2 = trace.correlate(b, a, normalization='gliding')
        assert numeq(c_ab.ydata, c_ba.ydata[::-1], 0.001)
        assert numeq(c_ab2.ydata, c_ba2.ydata[::-1], 0.001)

    def testNumpyCorrelate(self):
        primes = num.array(
            [1, 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31], dtype=num.int)
        n = 6
        lines = []
        have_flips, have_errs = False, False
        for ia in range(1, n+1):
            for ib in range(1, n+1):

                a = primes[:ia]
                b = primes[ia:ia+ib]

                for mode in 'full', 'valid', 'same':
                    c1 = trace.numpy_correlate_emulate(a, b, mode=mode)
                    c2 = trace.numpy_correlate_fixed(a, b, mode=mode)

                    if a.size < b.size:
                        sl = '<'
                    elif a.size > b.size:
                        sl = '>'
                    else:
                        sl = '='

                    assert c1.dtype == c2.dtype

                    sa = ''
                    if num.all(c1 == c1[::-1]):
                        sa = ' (ambig)'

                    if num.all(c1 == c2):
                        sr = 'ok' + sa
                    elif num.all(c1 == c2[::-1]):
                        have_flips = True
                        sr = 'flip'
                    else:
                        have_errs = True
                        sr = 'err, %s != %s' % (c1, c2)

                    kmin, kmax = trace.numpy_correlate_lag_range(
                        a, b, mode=mode)

                    lines.append(
                        'len(a) = %i, len(b) = %i, '
                        'kmin = %2i, kmax = %2i, (%s), mode = %-6s %s'
                        % (a.size, b.size, kmin, kmax, sl, mode+', ', sr))

        assert not have_flips and not have_errs, '\n'.join(lines)

    def testFFTCorrelate(self):
        for na in range(1, 20):
            for nb in range(1, 20):
                a1 = num.random.random(na)
                a2 = num.random.random(nb)

                t1 = trace.Trace(station='t1', ydata=a1)
                t2 = trace.Trace(station='t2', ydata=a2)

                for mode in 'full', 'same', 'valid':
                    c1 = trace.correlate(t1, t2, mode=mode)
                    c2 = trace.correlate(t1, t2, mode=mode, use_fft=True)
                    c1.set_station('c1')
                    c2.set_station('c2')
                    assert c1.get_ydata().size == c2.get_ydata().size
                    d = num.abs(c1.get_ydata() - c2.get_ydata())

                    if not num.all(d < 1e-5):
                        assert mode == 'same'

                        assert abs(abs(c1.tmin-c2.tmin) - 1.0) < 1e-5
                        assert abs(abs(c1.tmax-c2.tmax) - 1.0) < 1e-5

                        tmin = max(c1.tmin, c2.tmin)
                        tmax = min(c1.tmax, c2.tmax)
                        c1.chop(tmin, tmax, include_last=True)
                        c2.chop(tmin, tmax, include_last=True)

                        d = num.abs(c1.get_ydata() - c2.get_ydata())
                        assert(num.all(d < 1e-5))
                    else:
                        assert num.all(d < 1e-5)

    def testMovingSum(self):

        x = num.arange(5)
        assert numeq(
            trace.moving_sum(x, 3, mode='valid'),
            [3, 6, 9], 0.001)
        assert numeq(
            trace.moving_sum(x, 3, mode='full'),
            [0, 1, 3, 6, 9, 7, 4], 0.001)

    def testContinuousDownsample(self):

        y = num.random.random(1000)

        try:
            for (dt1, dt2) in [
                    (0.1, 1.0),
                    (0.2, 1.0),
                    (0.5, 1.0),
                    (0.2, 0.4),
                    (0.4, 1.2)]:

                for tadd in (0.0, 0.01, 0.2, 0.5, 0.7, 0.75):
                    a = trace.Trace(tmin=sometime+tadd, deltat=dt1, ydata=y)
                    bs = [trace.Trace(
                        location='b',
                        tmin=sometime+i*dt1*100+tadd,
                        deltat=dt1,
                        ydata=y[i*100:(i+1)*100]) for i in range(10)]

                    a.downsample_to(dt2, demean=False, snap=True)
                    c2s = []
                    downsampler = trace.co_downsample_to(
                        trace.co_list_append(c2s), dt2)

                    for b in bs:
                        c = downsampler.send(b)
                        c2s.append(c)

                    downsampler.close()
                    assert (round(c2s[0].tmin / dt2) * dt2 - c2s[0].tmin) \
                        / dt1 < 0.5001

        except trace.ScipyBug:
            raise unittest.SkipTest(
                'not supported on installed scipy version')

    def testEqualizeSamplingRates(self):
        y = num.random.random(1000)
        t1 = trace.Trace(tmin=0, ydata=y, deltat=0.01)
        t2 = trace.Trace(tmin=0, ydata=y, deltat=0.5)
        t1_new, t2_new = trace.equalize_sampling_rates(t1, t2)
        self.assertEqual(
            t1_new.deltat, t2_new.deltat,
            'Equalization of sampling rates failed: dt1=%s, dt2=%s'
            % (t1.deltat, t2.deltat))

    def testLxnorm(self):
        yref = num.array([0., 1.5, 2.3, 0., -1.])
        ytest = num.array([-1., 0.3, -0.3, 1., 0.2])
        m, n = trace.Lx_norm(ytest, yref, norm=1)
        self.assertEqual(m, 7., 'L1-norm: m is not 7., but %s' % str(m))
        self.assertEqual(n, 4.8, 'L1-norm: n is not 4.8, but %s' % str(n))

    def testMisfitOfSameTracesZero(self):
        y = num.random.random(10000)
        y -= max(y)*0.5
        t1 = trace.Trace(tmin=0, ydata=y, deltat=0.01)
        t2 = trace.Trace(tmin=0, ydata=y, deltat=0.01)
        # ttraces = [t2]
        fresponse = trace.FrequencyResponse()
        taper = trace.CosFader(xfade=2.)
        norms = [1, 2]
        domains = ['time_domain', 'frequency_domain', 'envelope', 'absolute']
        setups = [trace.MisfitSetup(
            norm=n,
            taper=taper,
            domain=domain,
            filter=fresponse) for domain in domains for n in norms]

        for setup in setups:
            m, n, tr, tc = t1.misfit(candidate=t2, setup=setup, debug=True)
            if isinstance(tr, trace.Trace):
                self.assertEqual(tr.tmin, tc.tmin)
                self.assertTrue(all(tr.get_ydata() == tc.get_ydata()))
            else:
                self.assertTrue(all(tc == tr))

            self.assertEqual(m, 0., 'misfit\'s m of equal traces is != 0')

    def testMisfitOfSameTracesDtDifferentShifted(self):
        test_file = common.test_data_file('1989.072.evt.mseed')

        p = pile.make_pile(test_file, show_progress=False)
        rt = p.all()[0]
        tt = rt.copy()

        # make downsampled, chopped copies:
        deltats = [0.5, 1.0, 2.0]
        tt.chop(tmin=rt.tmin+10, tmax=rt.tmax-15)
        tts = [tt.copy() for i in range(len(deltats))]
        [t.downsample_to(deltats[i]) for i, t in enumerate(tts)]

        # shift traces:
        t_shifts = [1.0, 0.49999, 0.5]
        for ts in t_shifts:
            tts_shifted = [t.copy() for t in tts]
            for x in tts_shifted:
                x.shift(ts)
            tts.extend(tts_shifted)

        a = rt.tmin
        d = rt.tmax
        b = a+(d-a)/10.
        c = d-(d-a)/10.

        taper = trace.CosTaper(a, b, c, d)
        fresponse = trace.FrequencyResponse()
        norms = [1, 2]
        domains = ['time_domain', 'frequency_domain', 'envelope', 'absolute']
        setups = [trace.MisfitSetup(
            norm=n,
            taper=taper,
            domain=domain,
            filter=fresponse) for domain in domains for n in norms]

        for cand in tts:
            for setup in setups:
                m, n = rt.misfit(candidate=cand, setup=setup)
                self.assertNotEqual(m, None, 'misfit\'s m is None')

    def testMisfitBox(self):

        ydata = num.zeros(9)
        ydata[3:6] = 1.0

        deltat = 0.5

        tr = trace.Trace(
            ydata=ydata,
            deltat=0.5, tmin=0.0)

        tshiftmin = - (tr.tmax - tr.tmin) * 2
        tshiftmax = (tr.tmax - tr.tmin) * 2

        tshifts = num.linspace(tshiftmin, tshiftmax,
                               int(round((tshiftmax-tshiftmin)/deltat))+1)

        candidates = []
        for tshift in tshifts:
            trc = tr.copy()
            trc.shift(tshift)
            candidates.append(trc)

        trace.MisfitSetup(
            norm=2,
            taper=trace.CosTaper(tr.tmin, tr.tmin, tr.tmax, tr.tmax),
            domain='time_domain')

    def testValidateFrequencyResponses(self):
        ttrace = trace.Trace(ydata=num.random.random(1000))
        inverse_eval = trace.InverseEvalresp(respfile='test.txt',
                                             trace=ttrace,
                                             target='vel')
        inverse_eval.validate()

        pzk_response = trace.PoleZeroResponse(zeros=[0+0j, 0+0j],
                                              poles=[1+0j, 2+0j],
                                              constant=10.)
        pzk_response.regularize()

    def testStretch(self):

        n = 100
        ydata = num.ones(n)
        ydata[n//2-10:n//2+10] = 2.0

        deltat = 0.01
        tr1 = trace.Trace(
            location='tr1',
            ydata=ydata,
            deltat=deltat,
            tmin=0.0)

        tr2 = tr1.copy()
        tr2.stretch(tr1.tmin-deltat, tr1.tmax+deltat)
        tr2.set_codes(location='tr2')

        tr3 = tr2.copy()
        tr3.stretch(tr1.tmin, tr1.tmax)
        tr3.set_codes(location='tr3')
        # trace.snuffle([tr1, tr2, tr3])

        assert numeq(tr1.ydata, tr3.ydata, 0.01)

    def benchmark_sinc(self):
        n = 10000000
        i_control = num.array([0., n-1], dtype=num.int64)
        t_control = num.array([0., n-1], dtype=num.float)
        s_in = num.zeros(n, dtype=num.float)
        s_out = num.zeros(n, dtype=num.float)
        tmin = 0.
        deltat = 1.0
        from pyrocko import signal_ext
        t1 = time.time()
        signal_ext.antidrift(i_control, t_control, s_in, tmin, deltat, s_out)
        t2 = time.time()
        print(t2 - t1)

    def test_transfer(self):
        n = 100
        ydata = num.random.random(n)
        deltat = 0.01
        tr1 = trace.Trace(
            location='tr1',
            ydata=ydata,
            deltat=deltat,
            tmin=0.0)
        tr2 = tr1.transfer()
        tr2.ydata += tr1.ydata.mean()
        assert numeq(tr1.ydata, tr2.ydata, 0.01)

    def test_muliply_taper(self):

        taper = trace.CosTaper(0., 1., 2., 3.)
        taper2 = trace.MultiplyTaper([taper, taper])

        y = num.ones(31)
        taper(y, 0., 0.1)
        taper(y, 0., 0.1)

        z = num.ones(31)
        taper2(z, 0., 0.1)

        assert numeq(y, z, 1e-6)
        assert taper.time_span() == taper2.time_span()

    def test_pickle(self):
        y = num.random.random(10000)
        t1 = trace.Trace(tmin=0, ydata=y, deltat=0.01)
        s = pickle.dumps(t1)
        t2 = pickle.loads(s)
        assert numeq(t1.ydata, t2.ydata, 0.0)

    def test_sta_lta_right(self):

        n = 100
        y = num.zeros(n)
        y[:] = 0.00001

        y[50:52] = num.linspace(0., 1., 2)
        y[52:] = 1.0

        tr = trace.Trace(
            '', 'X', '', '',
            ydata=y, deltat=1.0)

        for itlong in range(1, 20):
            for itshort in range(1, itlong):

                tlong = float(itlong)
                tshort = float(itshort)

                stalta = tr.copy()
                stalta.sta_lta_right(tshort, tlong)
                stalta.set_codes(location='stalta')

                ydata = num.zeros(n)

                sh = int(round(tshort / tr.deltat))
                ln = int(round(tlong / tr.deltat))

                for i in range(ln-sh, n-sh):
                    ydata[i] = (1.0 / sh * num.sum(y[i:i+sh])) / (
                        1.0 / ln * num.sum(y[i+sh-ln:i+sh])) \
                        * float(sh)/float(ln)

                stalta_ref = trace.Trace(
                    '', 'X', 'ref', '',
                    ydata=ydata, deltat=1.0)

                stalta_ref.chop(float(ln-sh), float(n-sh))

                assert numeq(stalta.ydata, stalta_ref.ydata, 1e-3)

    def test_trace_equal(self):
        ydata = num.arange(10)
        tr1 = trace.Trace(ydata=ydata)
        tr2 = trace.Trace(ydata=ydata)

        tr_set = set([tr1, tr2])

        assert len(tr_set), 2


if __name__ == "__main__":
    util.setup_logging('test_trace', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import unittest
from tempfile import mkdtemp
import shutil

from pyrocko import scenario, util, gf, gmtpy, config
from pyrocko.scenario import targets

util.force_dummy_progressbar = True

km = 1000.


def have_store(store_id):
    engine = gf.get_engine()
    try:
        engine.get_store(store_id)
        return True
    except gf.NoSuchStore:
        return False


def have_kite():
    try:
        import kite  # noqa
        return True
    except ImportError:
        return False


def have_srtm_credentials():
    if config.config().earthdata_credentials is None:
        return False
    return True


class ScenarioTestCase(unittest.TestCase):
    store_id = 'crust2_m5_hardtop_8Hz_fine'
    store_id_static = 'ak135_static'

    tempdirs = []

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            continue
            shutil.rmtree(d)

    @unittest.skipUnless(
        have_store(store_id),
        'GF Store "%s" is not available' % store_id)
    def test_scenario_waveforms(self):
        tempdir = mkdtemp(prefix='pyrocko-scenario')
        self.tempdirs.append(tempdir)

        vmin = 2500.

        generator = scenario.ScenarioGenerator(
            seed=20,
            center_lat=42.6,
            center_lon=13.3,
            radius=60*km,
            target_generators=[
                targets.WaveformGenerator(
                    store_id=ScenarioTestCase.store_id,
                    station_generator=targets.RandomStationGenerator(
                        nstations=5,
                        avoid_water=False),
                    noise_generator=targets.waveform.WhiteNoiseGenerator(),
                    seismogram_quantity='velocity'),
                ],
            source_generator=scenario.DCSourceGenerator(
                time_min=util.str_to_time('2017-01-01 00:00:00'),
                time_max=util.str_to_time('2017-01-01 02:00:00'),
                radius=10*km,
                depth_min=1*km,
                depth_max=10*km,
                magnitude_min=3.0,
                magnitude_max=4.5,
                strike=120.,
                dip=45.,
                rake=90.,
                perturbation_angle_std=15.,
                nevents=3)
        )

        def twin(source):
            tmin = source.time
            tmax = source.time + 100*km / vmin
            return tmin, tmax

        engine = gf.get_engine()
        generator.init_modelling(engine)

        ref_sources = generator.get_sources()
        ref_trs_list = []
        for source in ref_sources:
            trs = generator.get_waveforms(*twin(source))
            trs.sort(key=lambda tr: tr.nslc_id)
            ref_trs_list.append(trs)

        collection = scenario.ScenarioCollection(tempdir, engine)
        collection.add_scenario('one', generator)
        with self.assertRaises(scenario.ScenarioError):
            collection.add_scenario('one', generator)

        assert len(collection.list_scenarios()) == 1
        assert collection.list_scenarios()[0].scenario_id == 'one'

        s = collection.get_scenario('one')

        for ref_trs, source in zip(
                ref_trs_list,
                s.get_generator().get_sources()):

            trs = generator.get_waveforms(*twin(source))
            trs.sort(key=lambda tr: tr.nslc_id)
            self.assert_traces_almost_equal(trs, ref_trs)

        collection2 = scenario.ScenarioCollection(tempdir, engine)

        assert len(collection2.list_scenarios()) == 1
        assert collection2.list_scenarios()[0].scenario_id == 'one'

        s = collection2.get_scenario('one')

        for ref_trs, source in zip(
                ref_trs_list,
                s.get_generator().get_sources()):

            trs = generator.get_waveforms(*twin(source))
            trs.sort(key=lambda tr: tr.nslc_id)
            self.assert_traces_almost_equal(trs, ref_trs)

        tmin, tmax = s.get_time_range()
        s.ensure_waveforms(tmin, tmax)
        p = s.get_waveform_pile()

        for ref_trs, source in zip(
                ref_trs_list,
                s.get_generator().get_sources()):

            tmin, tmax = twin(source)
            trs = p.all(tmin=tmin, tmax=tmax, include_last=False)
            trs.sort(key=lambda tr: tr.nslc_id)
            self.assert_traces_almost_equal(trs, ref_trs)

    @unittest.skipUnless(
        have_kite(),
        'Kite is not available')
    @unittest.skipUnless(
        have_store(store_id),
        'GF Store "%s" is not available' % store_id)
    @unittest.skipUnless(
        have_store(store_id_static),
        'GF Store "%s" is not available' % store_id_static)
    def test_scenario_insar(self):
        tempdir = mkdtemp(prefix='pyrocko-scenario')
        self.tempdirs.append(tempdir)

        generator = scenario.ScenarioGenerator(
            seed=20,
            center_lat=42.6,
            center_lon=13.3,
            radius=60*km,
            target_generators=[
                targets.InSARGenerator(
                    resolution=(20, 20),
                    avoid_water=False,
                    noise_generator=targets.insar.AtmosphericNoiseGenerator(
                        amplitude=1e-5))
                ],
            source_generator=scenario.DCSourceGenerator(
                time_min=util.str_to_time('2017-01-01 00:00:00'),
                time_max=util.str_to_time('2017-01-01 02:00:00'),
                radius=10*km,
                depth_min=1*km,
                depth_max=10*km,
                magnitude_min=3.0,
                magnitude_max=4.5,
                strike=120.,
                dip=45.,
                rake=90.,
                perturbation_angle_std=15.,
                nevents=3)
        )

        engine = gf.get_engine()
        generator.init_modelling(engine)

        collection = scenario.ScenarioCollection(tempdir, engine)
        collection.add_scenario('insar', generator)

        s = collection.get_scenario('insar')
        s.ensure_insar_scenes()

    @unittest.skipUnless(
        have_store(store_id_static),
        'GF Store "%s" is not available' % store_id_static)
    def test_scenario_gnss(self):
        tempdir = mkdtemp(prefix='pyrocko-scenario')
        self.tempdirs.append(tempdir)

        generator = scenario.ScenarioGenerator(
            seed=20,
            center_lat=42.6,
            center_lon=13.3,
            radius=60*km,
            target_generators=[
                targets.GNSSCampaignGenerator(
                    station_generator=targets.RandomStationGenerator(
                        avoid_water=False,
                        channels=None))
                ],
            source_generator=scenario.DCSourceGenerator(
                time_min=util.str_to_time('2017-01-01 00:00:00'),
                time_max=util.str_to_time('2017-01-01 02:00:00'),
                radius=10*km,
                depth_min=1*km,
                depth_max=10*km,
                magnitude_min=3.0,
                magnitude_max=4.5,
                strike=120.,
                dip=45.,
                rake=90.,
                perturbation_angle_std=15.,
                nevents=3)
        )

        engine = gf.get_engine()
        generator.init_modelling(engine)

        collection = scenario.ScenarioCollection(tempdir, engine)
        collection.add_scenario('gnss', generator)

        s = collection.get_scenario('gnss')
        assert len(s.get_gnss_campaigns()) == 1

    @unittest.skipUnless(
        have_kite(),
        'Kite is not available')
    @unittest.skipUnless(
        have_store(store_id),
        'GF Store "%s" is not available' % store_id)
    def test_scenario_combinations(self):

        generator = scenario.ScenarioGenerator(
            seed=20,
            center_lat=42.6,
            center_lon=13.3,
            radius=60*km,
            target_generators=[
                targets.RandomStationGenerator(
                    avoid_water=False,
                    nstations=5),
                targets.WaveformGenerator(
                    store_id=ScenarioTestCase.store_id,
                    station_generator=targets.RandomStationGenerator(
                        avoid_water=False),
                    noise_generator=targets.waveform.WhiteNoiseGenerator(),
                    seismogram_quantity='velocity'),
                targets.InSARGenerator(
                    resolution=(20, 20),
                    avoid_water=False,
                    noise_generator=targets.insar.AtmosphericNoiseGenerator(
                        amplitude=1e-5)),
                targets.GNSSCampaignGenerator(
                    station_generator=targets.RandomStationGenerator(
                        avoid_water=False,
                        channels=None))
                ],
            source_generator=scenario.DCSourceGenerator(
                time_min=util.str_to_time('2017-01-01 00:00:00'),
                time_max=util.str_to_time('2017-01-01 02:00:00'),
                radius=10*km,
                depth_min=1*km,
                depth_max=10*km,
                magnitude_min=3.0,
                magnitude_max=4.5,
                strike=120.,
                dip=45.,
                rake=90.,
                perturbation_angle_std=15.,
                nevents=3)
        )

        engine = gf.get_engine()
        generator.init_modelling(engine)

        for src in scenario.sources.AVAILABLE_SOURCES:
            generator.source_generator = src(
                time_min=util.str_to_time('2017-01-01 00:00:00'),
                time_max=util.str_to_time('2017-01-01 02:00:00'),
                radius=1*km,
                depth_min=1.5*km,
                depth_max=5*km,
                magnitude_min=3.0,
                magnitude_max=4.5)
            generator.source_generator.update_hierarchy(generator)

            generator.get_stations()
            generator.get_waveforms()
            generator.get_insar_scenes()
            generator.get_gnss_campaigns()

    @unittest.skipUnless(
        have_store(store_id_static),
        'GF Store "%s" is not available' % store_id_static)
    @unittest.skipUnless(
        gmtpy.have_gmt(), 'GMT not available')
    @unittest.skipUnless(
        have_srtm_credentials(),
        'No Earthdata credentials in config.')
    def test_scenario_map(self):
        tempdir = mkdtemp(prefix='pyrocko-scenario')
        self.tempdirs.append(tempdir)

        generator = scenario.ScenarioGenerator(
            seed=20,
            center_lat=42.6,
            center_lon=13.3,
            radius=60*km,
            target_generators=[
                targets.RandomStationGenerator(
                    avoid_water=False,
                    nstations=5),
                targets.WaveformGenerator(
                    store_id=ScenarioTestCase.store_id,
                    station_generator=targets.RandomStationGenerator(
                        avoid_water=False),
                    noise_generator=targets.waveform.WhiteNoiseGenerator(),
                    seismogram_quantity='velocity'),
                targets.InSARGenerator(
                    resolution=(20, 20),
                    avoid_water=False,
                    noise_generator=targets.insar.AtmosphericNoiseGenerator(
                        amplitude=1e-5)),
                targets.GNSSCampaignGenerator(
                    station_generator=targets.RandomStationGenerator(
                        avoid_water=False,
                        channels=None))
                ],
            source_generator=scenario.DCSourceGenerator(
                time_min=util.str_to_time('2017-01-01 00:00:00'),
                time_max=util.str_to_time('2017-01-01 02:00:00'),
                radius=10*km,
                depth_min=1*km,
                depth_max=10*km,
                magnitude_min=3.0,
                magnitude_max=4.5,
                strike=120.,
                dip=45.,
                rake=90.,
                perturbation_angle_std=15.,
                nevents=3)
        )

        engine = gf.get_engine()

        collection = scenario.ScenarioCollection(tempdir, engine)
        collection.add_scenario('plot', generator)

        s = collection.get_scenario('plot')
        s.get_map()

    def assert_traces_almost_equal(self, trs1, trs2):
        assert len(trs1) == len(trs2)
        for (tr1, tr2) in zip(trs1, trs2):
            tr1.assert_almost_equal(tr2)


if __name__ == '__main__':
    util.setup_logging('test_scenario', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import range, next

import time
import sys
import random
import math
import unittest
import logging
import numpy as num
import shutil
from tempfile import mkdtemp

from pyrocko import guts
from pyrocko import gf, util, cake, ahfullgreen, trace
from pyrocko.fomosto import ahfullgreen as fomosto_ahfullgreen

from .common import Benchmark

assert_ae = num.testing.assert_almost_equal


logger = logging.getLogger('pyrocko.test.test_gf')
benchmark = Benchmark()

local_stores = gf.LocalEngine(use_config=True).get_store_ids()

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.


def arr(x):
    return num.asarray(x, dtype=num.float)


def numeq(a, b, eps):
    return (num.all(num.asarray(a).shape == num.asarray(b).shape and
            num.abs(num.asarray(a) - num.asarray(b)) < eps))


def _make_traces_homogeneous(
        dsource, receiver, material, deltat, net, sta, loc):

    comps = ['displacement.n', 'displacement.e', 'displacement.d']
    npad = 120

    dists = dsource.distances_to(receiver)

    dists = dsource.distances_to(receiver)
    azis, _ = dsource.azibazis_to(receiver)

    norths = dists * num.cos(azis*d2r)
    easts = dists * num.sin(azis*d2r)

    ddepths = receiver.depth - dsource.depths

    d3ds = num.sqrt(norths**2 + easts**2 + ddepths**2)
    times = dsource.times
    toff = math.floor(num.min(times) / deltat) * deltat

    tmin = num.min(
        (num.floor(times + d3ds / material.vp / deltat) - npad) * deltat)
    tmax = num.max(
        (num.ceil(times + d3ds / material.vs / deltat) + npad) * deltat)

    ns = int(round((tmax - tmin) / deltat))

    outx = num.zeros(ns)
    outy = num.zeros(ns)
    outz = num.zeros(ns)

    traces = []
    for ielement in range(dsource.nelements):
        x = (norths[ielement], easts[ielement], ddepths[ielement])

        if isinstance(dsource, gf.DiscretizedMTSource):
            m6 = dsource.m6s[ielement, :]
            f = (0., 0., 0.)
        elif isinstance(dsource, gf.DiscretizedSFSource):
            m6 = (0., 0., 0., 0., 0., 0.)
            f = dsource.forces[ielement, :]
        elif isinstance(dsource, gf.DiscretizedExplosionSource):
            m0 = dsource.m0s[ielement]
            m6 = (m0, m0, m0, 0., 0., 0.)
            f = (0., 0., 0.)
        else:
            assert False

        ahfullgreen.add_seismogram(
            material.vp, material.vs, material.rho, material.qp, material.qs,
            x, f, m6, 'displacement',
            deltat, tmin - (times[ielement] - toff), outx, outy, outz,
            stf=ahfullgreen.Impulse())

    for comp, o in zip(comps, (outx, outy, outz)):
        tr = trace.Trace(
            net, sta, loc, comp,
            tmin=tmin + toff, ydata=o, deltat=deltat)

        traces.append(tr)

    return traces


class PulseConfig(guts.Object):
    fwhm = guts.Float.T(default=0.02)
    velocity = guts.Float.T(default=1000.)

    def evaluate(self, distance, t):
        tarr = distance / self.velocity
        denom = self.fwhm / (2.*math.sqrt(math.log(2.)))
        return num.exp(-((t-tarr)/denom)**2)


class GFTestCase(unittest.TestCase):

    tempdirs = []

    if sys.version_info < (2, 7):
        from contextlib import contextmanager

        @contextmanager
        def assertRaises(self, exc):

            gotit = False
            try:
                yield None
            except exc:
                gotit = True

            assert gotit, 'expected to get a %s exception' % exc

        def assertIsNone(self, value):
            assert value is None, 'expected None but got %s' % value

    @classmethod
    def setUpClass(cls):
        cls.pulse_store_dir = None
        cls.regional_ttt_store_dir = None
        cls.benchmark_store_dir = None
        cls._dummy_store = None

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def create(self, deltat=1.0, nrecords=10):
        d = mkdtemp(prefix='gfstore')
        store = gf.BaseStore.create(d, deltat, nrecords, force=True)

        store = gf.BaseStore(d, mode='w')
        for i in range(nrecords):
            data = num.asarray(num.random.random(random.randint(0, 7)),
                               dtype=gf.gf_dtype)

            tr = gf.GFTrace(data=data, itmin=1+i)
            store.put(i, tr)

        store.close()
        self.tempdirs.append(d)
        return d

    def get_pulse_store_dir(self):
        if self.pulse_store_dir is None:
            self.pulse_store_dir = self._create_pulse_store()

        return self.pulse_store_dir

    def get_regional_ttt_store_dir(self):
        if self.regional_ttt_store_dir is None:
            self.regional_ttt_store_dir = self._create_regional_ttt_store()

        return self.regional_ttt_store_dir

    def get_benchmark_store_dir(self):
        if self.benchmark_store_dir is None:
            self.benchmark_store_dir = self._create_benchmark_store()

        return self.benchmark_store_dir

    def _create_benchmark_store(self):
        conf = gf.ConfigTypeA(
            id='benchmark_store',
            source_depth_min=0.,
            source_depth_max=2.,
            source_depth_delta=1.,
            distance_min=1.0,
            distance_max=5001.0,
            distance_delta=5.0,
            sample_rate=2.0,
            ncomponents=5)

        deltat = 1.0/conf.sample_rate

        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)

        gf.Store.create(store_dir, config=conf)
        store = gf.Store(store_dir, 'w')
        for args in conf.iter_nodes():
            nsamples = int(round(args[1]))
            data = num.ones(nsamples)
            itmin = int(round(args[1]))
            tr = gf.GFTrace(data=data, itmin=itmin, deltat=deltat)
            store.put(args, tr)

        store.close()
        return store_dir

    def _create_regional_ttt_store(self):

        conf = gf.ConfigTypeA(
            id='empty_regional',
            source_depth_min=0.,
            source_depth_max=20*km,
            source_depth_delta=10*km,
            distance_min=1000*km,
            distance_max=2000*km,
            distance_delta=10*km,
            sample_rate=2.0,
            ncomponents=10,
            earthmodel_1d=cake.load_model(),
            tabulated_phases=[
                gf.TPDef(id=id, definition=defi) for (id, defi) in [
                    ('depthp', 'p'),
                    ('pS', 'pS'),
                    ('P', 'P'),
                    ('S', 'S')
                ]
            ])

        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)

        gf.Store.create(store_dir, config=conf)
        store = gf.Store(store_dir)
        store.make_ttt()

        store.close()
        return store_dir

    def _create_pulse_store(self):

        conf = gf.ConfigTypeB(
            id='pulse',
            receiver_depth_min=0.,
            receiver_depth_max=10.,
            receiver_depth_delta=10.,
            source_depth_min=0.,
            source_depth_max=1000.,
            source_depth_delta=10.,
            distance_min=10.,
            distance_max=1000.,
            distance_delta=10.,
            sample_rate=200.,
            ncomponents=2.,
            component_scheme='elastic2')

        pulse = PulseConfig()

        # fnyq_spatial = pulse.velocity / math.sqrt(conf.distance_delta**2 +
        #                                       conf.source_depth_delta**2)

        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)

        gf.Store.create(store_dir, config=conf, force=True,
                        extra={'pulse': pulse})

        deltat = conf.deltat

        store = gf.Store(store_dir, mode='w')
        for args in store.config.iter_nodes(level=-1):

            rdepth, sdepth, surfdist = args
            dist = math.sqrt((rdepth-sdepth)**2 + surfdist**2)

            tarr = dist / pulse.velocity

            tmin = tarr - 5 * pulse.fwhm
            tmax = tarr + 5 * pulse.fwhm
            itmin = int(num.floor(tmin/deltat))
            itmax = int(num.ceil(tmax/deltat))
            tmin = itmin * deltat
            tmax = itmax * deltat
            nsamples = itmax - itmin + 1

            t = tmin + num.arange(nsamples) * deltat

            data = pulse.evaluate(dist, t)

            phi = math.atan2(rdepth-sdepth, surfdist)
            data = [data*math.cos(phi), data*math.sin(phi)]
            for icomponent, data in enumerate(data):
                is_zero = num.all(data == 0.0)

                tr = gf.GFTrace(data=data, itmin=itmin, deltat=deltat,
                                is_zero=is_zero)

                store.put(args + (icomponent,), tr)

        store.close()
        return store_dir

    def test_get_spans(self):
        nrecords = 8
        random.seed(0)
        num.random.seed(0)

        store = gf.BaseStore(self.create(nrecords=nrecords))
        for i in range(nrecords):
            for deci in (1, 2, 3, 4):
                tr = store.get(i, decimate=deci)
                itmin, itmax = store.get_span(i, decimate=deci)
                self.assertEqual(tr.itmin, itmin)
                self.assertEqual(tr.data.size, itmax-itmin + 1)

        store.close()

    def test_get_shear_moduli(self):
        store_dir = self.get_regional_ttt_store_dir()
        store = gf.Store(store_dir)

        sample_points = num.empty((20, 3))
        sample_points[:, 2] = num.linspace(
            0, store.config.coords[0].max(), 20)

        for interp in ('nearest_neighbor', 'multilinear'):
            store.config.get_shear_moduli(
                lat=0., lon=0.,
                points=sample_points,
                interpolation=interp)

    def test_partial_get(self):
        nrecords = 8
        random.seed(0)
        num.random.seed(0)

        store = gf.BaseStore(self.create(nrecords=nrecords))
        for deci in (1, 2, 3, 4):
            for i in range(0, nrecords):
                tra = store.get(i, decimate=deci, implementation='c')
                trb = store.get(i, decimate=deci, implementation='python')
                self.assertEqual(tra.itmin, trb.itmin)
                self.assertEqual(tra.data.size, trb.data.size)
                assert_ae(tra.data, trb.data)
                assert_ae(tra.begin_value, trb.begin_value)
                assert_ae(tra.end_value, trb.end_value)

                tr = tra

                itmin_gf, nsamples_gf = tr.itmin, tr.data.size
                for itmin in range(itmin_gf - nsamples_gf,
                                   itmin_gf + nsamples_gf+1):

                    for nsamples in range(0, nsamples_gf*3):
                        for implementation in ['c', 'python']:
                            tr2 = store.get(i, itmin, nsamples, decimate=deci)
                            self.assertEqual(tr2.itmin, max(tr.itmin, itmin))
                            self.assertEqual(tr2.itmin + tr2.data.size,
                                             max(min(tr.itmin + tr.data.size,
                                                     itmin + nsamples),
                                                 tr2.itmin + tr2.data.size))

                            ilo = max(tr.itmin, tr2.itmin)
                            ihi = min(tr.itmin+tr.data.size,
                                      tr2.itmin+tr2.data.size)

                            a = tr.data[ilo-tr.itmin:ihi-tr.itmin]
                            b = tr2.data[ilo-tr2.itmin:ihi-tr2.itmin]

                            self.assertTrue(numeq(a, b, 0.001))

        store.close()

    def test_sum(self):

        nrecords = 8
        random.seed(0)
        num.random.seed(0)

        store = gf.BaseStore(self.create(nrecords=nrecords))

        from pyrocko.gf import store_ext
        store.open()

        store_ext.store_mapping_init(
            store.cstore, 'type_0',
            arr([0]), arr([nrecords-1]), arr([1]),
            num.array([nrecords], dtype=num.uint64),
            1)

        for deci in (1, 2, 3, 4):
            for i in range(300):
                n = random.randint(0, 5)
                indices = num.random.randint(nrecords, size=n)
                weights = num.random.random(n)
                shifts = num.random.random(n)*nrecords
                shifts[::2] = num.round(shifts[::2])

                for itmin, nsamples in [(None, None),
                                        (random.randint(0, nrecords),
                                         random.randint(0, nrecords))]:

                    a = store.sum(
                        indices, shifts, weights,
                        itmin=itmin,
                        nsamples=nsamples,
                        decimate=deci)

                    b = store.sum(
                        indices, shifts, weights,
                        itmin=itmin,
                        nsamples=nsamples,
                        decimate=deci,
                        implementation='alternative')

                    c = store.sum(
                        indices, shifts, weights,
                        itmin=itmin,
                        nsamples=nsamples,
                        decimate=deci,
                        implementation='reference')

                    self.assertEqual(a.itmin, c.itmin)
                    num.testing.assert_array_almost_equal(
                        a.data, c.data, 2)

                    self.assertEqual(b.itmin, c.itmin)
                    num.testing.assert_array_almost_equal(
                        b.data, c.data, 2)

                    if deci != 1:
                        continue

                    nthreads = 1
                    source_coords = num.zeros((n, 5))
                    source_coords[:, 4] = indices
                    receiver_coords = num.zeros((1, 5))
                    source_terms = num.zeros((n, 1))
                    source_terms[:, 0] = weights
                    results = store_ext.store_calc_timeseries(
                        store.cstore,
                        source_coords,
                        source_terms,
                        shifts,
                        receiver_coords,
                        'dummy',
                        'nearest_neighbor',
                        num.array(
                            [itmin if itmin is not None else 0],
                            dtype=num.int32),
                        num.array(
                            [nsamples if nsamples is not None else -1],
                            dtype=num.int32),
                        nthreads)

                    d = gf.GFTrace(*results[0][:2])
                    self.assertEqual(a.itmin, d.itmin)
                    num.testing.assert_array_almost_equal(
                        a.data, d.data, 2)

        store.close()

    def _test_sum_statics(self):

        nrecords = 20

        store = gf.BaseStore(self.create(nrecords=nrecords))
        for i in range(5):
            n = random.randint(1, 10)
            indices = num.random.randint(nrecords, size=n).astype(num.uint64)
            weights = num.random.random(n).astype(num.float32)
            shifts = (num.random.random(n)*nrecords/4).astype(num.float32)
            it = random.randint(0, 5)

            dyn = store.sum(
                indices, shifts, weights,
                itmin=it,
                nsamples=1,
                decimate=1,
                implementation='c',
                optimization='enable')

            sta = store.sum_statics(
                indices, shifts, weights,
                it, 1,
                nthreads=1)

            if len(dyn.data) > 0:
                num.testing.assert_array_almost_equal(
                    dyn.data[-1], sta[0], 5)

        store.close()

    def test_store_dir_type(self):
        with self.assertRaises(TypeError):
            gf.LocalEngine(store_dirs='dummy')

    def test_pulse(self):
        store_dir = self.get_pulse_store_dir()

        engine = gf.LocalEngine(store_dirs=[store_dir])

        sources = [
            gf.ExplosionSource(
                time=0.0,
                depth=depth,
                moment=moment)

            for moment in (1.0, 2.0, 3.0) for depth in [100., 200., 300.]
        ]

        targets = [
            gf.Target(
                codes=('', 'STA', '', component),
                north_shift=500.,
                east_shift=0.)

            for component in 'ZNE'
        ]

        pulse = engine.get_store_extra(None, 'pulse')
        store = engine.get_store('pulse')

        response = engine.process(sources=sources, targets=targets)
        for source, target, tr in response.iter_results():
            t = tr.get_xdata()

            dist = math.sqrt((source.depth - target.depth)**2 +
                             source.distance_to(target)**2)

            data = pulse.evaluate(dist, t-source.time)

            phi = math.atan2((source.depth - target.depth),
                             source.distance_to(target)) * r2d

            azi, bazi = source.azibazi_to(target)

            data *= source.get_moment(store) * math.sqrt(2./3.)

            if tr.channel.endswith('N'):
                data *= math.cos(phi*d2r) * math.cos(azi*d2r)
            elif tr.channel.endswith('E'):
                data *= math.cos(phi*d2r) * math.sin(azi*d2r)
            elif tr.channel.endswith('Z'):
                data *= math.sin(phi*d2r)

            tr2 = tr.copy(data=False)
            tr2.set_ydata(data)
            tr2.set_codes(location='X')

            num.testing.assert_almost_equal(data, tr.ydata, 2)

    @unittest.skip('')
    def test_pulse_decimate(self):
        store_dir = self.get_pulse_store_dir()

        store = gf.Store(store_dir)
        store.make_decimated(2)

        engine = gf.LocalEngine(store_dirs=[store_dir])
        # pulse = engine.get_store_extra(None, 'pulse')

        source = gf.ExplosionSource(
            time=0.0,
            depth=100.,
            moment=1.0)

        targets = [
            gf.Target(
                codes=('', 'STA', '%s' % sample_rate, component),
                sample_rate=sample_rate,
                north_shift=500.,
                east_shift=0.)

            for component in 'N'
            for sample_rate in [None, store.config.sample_rate / 2.0]
        ]

        response = engine.process(source, targets)

        trs = []
        for source, target, tr in response.iter_results():
            tr.extend(0., 1.)
            if target.sample_rate is None:
                tr.downsample_to(2./store.config.sample_rate, snap=True)

            trs.append(tr)

        tmin = max(tr.tmin for tr in trs)
        tmax = min(tr.tmax for tr in trs)

        for tr in trs:
            tr.chop(tmin, tmax)

        num.testing.assert_almost_equal(
            trs[0].ydata, trs[1].ydata, 2)

    def test_stf_pre_post(self):
        store_dir = self.get_pulse_store_dir()
        engine = gf.LocalEngine(store_dirs=[store_dir])
        store = engine.get_store('pulse')

        for duration in [0., 0.05, 0.1]:
            trs = []
            for mode in ['pre', 'post']:
                source = gf.ExplosionSource(
                    time=store.config.deltat * 0.5,
                    depth=200.,
                    moment=1.0,
                    stf=gf.BoxcarSTF(duration=duration),
                    stf_mode=mode)

                target = gf.Target(
                    codes=('', 'STA', '', 'Z'),
                    north_shift=500.,
                    east_shift=0.,
                    store_id='pulse')

                xtrs = engine.process(source, target).pyrocko_traces()
                for tr in xtrs:
                    tr.set_codes(location='%3.1f_%s' % (duration, mode))
                    trs.append(tr)

            tmin = max(tr.tmin for tr in trs)
            tmax = min(tr.tmax for tr in trs)
            for tr in trs:
                tr.chop(tmin, tmax)

            amax = max(num.max(num.abs(tr.ydata)) for tr in trs)
            perc = num.max(num.abs(trs[0].ydata - trs[1].ydata) / amax) * 100.
            if perc > 0.1:
                logger.warning(
                    'test_stf_pre_post: max difference of %.1f %%' % perc)

    def test_target_source_timing(self):
        store_dir = self.get_pulse_store_dir()
        engine = gf.LocalEngine(store_dirs=[store_dir])

        for stime in [0., -160000., time.time()]:
            source = gf.ExplosionSource(
                        depth=200.,
                        magnitude=4.,
                        time=stime)

            targets = [
                gf.Target(
                    codes=('', 'STA', '', component),
                    north_shift=500.,
                    tmin=source.time-300.,
                    tmax=source.time+300.,
                    east_shift=500.)

                for component in 'ZNE'
            ]

            response = engine.process(source, targets)
            synthetic_traces = response.pyrocko_traces()
            data = num.zeros(num.shape(synthetic_traces[0].ydata))
            for tr in synthetic_traces:
                data += tr.ydata

            sum_data = num.sum(abs(tr.ydata))
            assert sum_data > 1.0

    def benchmark_get(self):
        store_dir = self.get_benchmark_store_dir()

        import pylab as lab
        for implementation in ('c', 'python'):
            store = gf.Store(store_dir, use_memmap=True)
            for nrepeats in (1, 2):
                data = []
                for distance in store.config.coords[1]:
                    sdepths = num.repeat(store.config.coords[0],
                                         store.config.ncomponents)
                    t = time.time()
                    for repeat in range(nrepeats):
                        for sdepth in sdepths:
                            for icomp in range(1):
                                store.get(
                                    (sdepth, distance, icomp),
                                    implementation=implementation)

                    tnew = time.time()
                    data.append((distance, tnew - t))

                if nrepeats != 1:
                    d, t1 = num.array(data, dtype=num.float).T
                    nread = nrepeats * store.config.ns[0]
                    smmap = implementation
                    label = 'nrepeats %i, impl %s' % (nrepeats, smmap)
                    print(label, num.mean(nread/t1))

                    lab.plot(d, nread/t1, label=label)

        lab.legend()
        lab.show()

    def benchmark_sum(self):

        store_dir = self.get_benchmark_store_dir()

        import pylab as lab
        for implementation in ('c', 'python'):
            store = gf.Store(store_dir, use_memmap=True)
            for weight in (0.0, 1.0):
                for nrepeats in (1, 2):
                    data = []
                    for distance in store.config.coords[1]:
                        n = store.config.ncomponents*store.config.ns[0]
                        sdepths = num.repeat(store.config.coords[0],
                                             store.config.ncomponents)
                        distances = num.repeat([distance], n)
                        comps = num.tile(store.config.coords[2],
                                         store.config.ns[0])
                        args = (sdepths, distances, comps)
                        weights = num.repeat([weight], n)
                        delays = num.arange(n, dtype=num.float) \
                            * store.config.deltat * 0.5

                        t = time.time()

                        for repeat in range(nrepeats):
                            store.sum(args, delays, weights,
                                      implementation=implementation)

                        tnew = time.time()

                        data.append(((distance-store.config.distance_min)+1,
                                     tnew - t))

                    if nrepeats != 1:
                        d, t1 = num.array(data, dtype=num.float).T
                        nread = nrepeats * store.config.ns[0] \
                            * store.config.ncomponents
                        label = 'nrepeats %i, weight %g, impl %s' % (
                            nrepeats, weight, implementation)
                        print(label, num.mean(nread/t1))

                        lab.plot(d, nread/t1, label=label)

        lab.legend()
        lab.show()

    def test_optimization(self):
        store_dir = self.get_pulse_store_dir()
        engine = gf.LocalEngine(store_dirs=[store_dir])

        sources = [
            gf.RectangularExplosionSource(
                time=0.0025,
                depth=depth,
                moment=1.0,
                length=100.,
                width=0.,
                nucleation_x=-1)

            for depth in [100., 200., 300.]
        ]

        targetss = [
            [
                gf.Target(
                    codes=('', 'STA', opt, component),
                    north_shift=500.,
                    east_shift=125.,
                    depth=depth,
                    interpolation='multilinear',
                    optimization=opt)

                for component in 'ZNE' for depth in [0., 5., 10]]
            for opt in ('disable', 'enable')
        ]

        resps = [engine.process(sources, targets) for targets in targetss]

        iters = [resp.iter_results() for resp in resps]
        for i in range(len(sources) * len(targetss[0])):
            s1, t1, tr1 = next(iters[0])
            s2, t2, tr2 = next(iters[1])
            self.assertEqual(tr1.data_len(), tr2.data_len())
            self.assertEqual(tr1.tmin, tr2.tmin)
            self.assertTrue(numeq(tr1.ydata, tr2.ydata, 0.0001))

    def test_timing_defs(self):

        for s, d in [
                ('100.0', dict(offset=100)),
                ('cake: P ', dict(offset=0.0, phases=['cake:P'])),
                ('iaspei: Pdiff', dict(
                    phases=['iaspei:Pdiff'])),
                ('{iaspei: Pdiff}-50', dict(
                    phases=['iaspei:Pdiff'], offset=-50.)),
                ('{cake: P | iaspei: Pdiff}-10', dict(
                    offset=-10.,
                    phases=['cake:P', 'iaspei:Pdiff'])),
                ('first{cake: p | cake: P}', dict(
                    phases=['cake:p', 'cake:P'], select='first')),
                ('first(p|P)-10', dict(
                    phases=['stored:p', 'stored:P'], select='first',
                    offset=-10.)),
                ('{stored:begin}-50', dict(
                    phases=['stored:begin'], offset=-50.))]:
            t = gf.Timing(s)

            if 'phases' in d:
                for a, b in zip(d['phases'], t.phase_defs):
                    self.assertEqual(a, b)

            if 'offset' in d:
                self.assertTrue(numeq(d['offset'], t.offset, 0.001))

            if 'select' in d:
                self.assertEqual(d['select'], t.select)

    def test_timing(self):
        store_dir = self.get_regional_ttt_store_dir()

        store = gf.Store(store_dir)

        args = (10*km, 1500*km)
        assert(store.t('P', args) is not None)
        self.assertEqual(store.t('last(S|P)', args), store.t('S', args))
        self.assertEqual(store.t('(S|P)', args), store.t('S', args))
        self.assertEqual(store.t('(P|S)', args), store.t('P', args))
        self.assertEqual(store.t('first(S|P)', args), store.t('P', args))

        with self.assertRaises(gf.NoSuchPhase):
            store.t('nonexistant', args)

        with self.assertRaises(AssertionError):
            store.t('P', (10*km,))

        with self.assertRaises(gf.OutOfBounds):
            print(store.t('P', (10*km, 5000*km)))

        with self.assertRaises(gf.OutOfBounds):
            print(store.t('P', (30*km, 1500*km)))

    def test_timing_new_syntax(self):
        store_dir = self.get_regional_ttt_store_dir()

        store = gf.Store(store_dir)

        args = (10*km, 1500*km)

        assert numeq(store.t('stored:P', args), store.t('cake:P', args), 0.1)
        assert numeq(store.t('vel_surface:15', args), 100., 0.1)
        assert numeq(store.t('+0.1S', args), 150., 0.1)
        assert numeq(
            store.t('{stored:P}+0.1S', args),
            store.t('{cake:P}', args) + store.t('{vel_surface:10}', args),
            0.1)

    def dummy_store(self):
        if self._dummy_store is None:

            conf = gf.ConfigTypeA(
                id='empty_regional',
                source_depth_min=0.,
                source_depth_max=20*km,
                source_depth_delta=1*km,
                distance_min=1*km,
                distance_max=2000*km,
                distance_delta=1*km,
                sample_rate=2.0,
                ncomponents=10)

            store_dir = mkdtemp(prefix='gfstore')
            self.tempdirs.append(store_dir)

            gf.Store.create(store_dir, config=conf)
            self._dummy_store = gf.Store(store_dir)

        return self._dummy_store

    def test_make_params(self):
        from pyrocko.gf import store_ext
        benchmark.show_factor = True

        def test_make_params_bench(store, dim, ntargets, interpolation,
                                   nthreads):
            source = gf.RectangularSource(
                lat=0., lon=0.,
                depth=10*km, north_shift=0.1, east_shift=0.1,
                width=dim, length=dim)

            targets = [gf.Target(
                lat=random.random()*10.,
                lon=random.random()*10.,
                north_shift=0.1,
                east_shift=0.1) for x in range(ntargets)]

            dsource = source.discretize_basesource(store, targets[0])
            source_coords_arr = dsource.coords5()
            mts_arr = dsource.m6s

            receiver_coords_arr = num.empty((len(targets), 5))
            for itarget, target in enumerate(targets):
                receiver = target.receiver(store)
                receiver_coords_arr[itarget, :] = \
                    [receiver.lat, receiver.lon, receiver.north_shift,
                     receiver.east_shift, receiver.depth]
            ns = mts_arr.shape[0]
            label = '_ns%04d_nt%04d_%s_np%02d' % (
                ns,
                len(targets),
                interpolation,
                nthreads)

            @benchmark.labeled('c%s' % label)
            def make_param_c():
                return store_ext.make_sum_params(
                    store.cstore,
                    source_coords_arr,
                    mts_arr,
                    receiver_coords_arr,
                    'elastic10',
                    interpolation, nthreads)

            @benchmark.labeled('p%s' % label)
            def make_param_python():
                weights_c = []
                irecords_c = []
                for itar, target in enumerate(targets):
                    receiver = target.receiver(store)
                    dsource = source.discretize_basesource(store, target)

                    for i, (component, args, delays, weights) in \
                            enumerate(store.config.make_sum_params(
                                dsource, receiver)):
                        if len(weights_c) <= i:
                            weights_c.append([])
                            irecords_c.append([])

                        if interpolation == 'nearest_neighbor':
                            irecords = num.array(store.config.irecords(*args))
                            weights = num.array(weights)
                        else:
                            assert interpolation == 'multilinear'
                            irecords, ip_weights =\
                                store.config.vicinities(*args)
                            neach = irecords.size // args[0].size
                            weights = num.repeat(weights, neach) * ip_weights
                            delays = num.repeat(delays, neach)

                        weights_c[i].append(weights)
                        irecords_c[i].append(irecords)
                for c in range(len(weights_c)):
                    weights_c[c] = num.concatenate([w for w in weights_c[c]])
                    irecords_c[c] = num.concatenate([ir for
                                                     ir in irecords_c[c]])

                return list(zip(weights_c, irecords_c))

            rc = make_param_c()
            rp = make_param_python()

            logger.info(benchmark.__str__(header=False))
            benchmark.clear()

            # Comparing the results
            if isinstance(store.config, gf.meta.ConfigTypeA):
                idim = 4
            elif isinstance(store.config, gf.meta.ConfigTypeB):
                idim = 8
            if interpolation == 'nearest_neighbor':
                idim = 1

            nsummands_scheme = [5, 5, 3]  # elastic8
            nsummands_scheme = [6, 6, 4]  # elastic10
            for i, nsummands in enumerate(nsummands_scheme):
                for r in [0, 1]:
                    r_c = rc[i][r]
                    r_p = rp[i][r].reshape(ntargets, nsummands, ns*idim)
                    r_p = num.transpose(r_p, axes=[0, 2, 1])

                    num.testing.assert_almost_equal(r_c, r_p.flatten())

        '''
        Testing loop
        '''
        # dims = [2*km, 5*km, 8*km, 16*km]
        # ntargets = [10, 100, 1000]

        # dims = [16*km]
        # ntargets = [1000]
        dims = [2*km, 5*km]
        ntargets = [10, 100]

        store = self.dummy_store()
        store.open()

        for interpolation in ['multilinear', 'nearest_neighbor']:
            for d in dims:
                for nt in ntargets:
                    for nthreads in [1, 2]:
                        test_make_params_bench(
                            store, d, nt, interpolation, nthreads)

        benchmark.show_factor = False

    @unittest.skipIf('global_2s' not in local_stores,
                     'depends on store global_2s')
    def test_calc_timeseries(self):
        from pyrocko.gf import store_ext
        benchmark.show_factor = True

        engine = gf.LocalEngine(use_config=True)

        def test_timeseries(
                store, source, targets, interpolation, nthreads,
                random_itmin=False, random_nsamples=False):

            ntargets = len(targets)
            dsource = source.discretize_basesource(store, targets[0])
            source_coords_arr = dsource.coords5()
            scheme = store.config.component_scheme
            source_terms = dsource.get_source_terms(scheme)

            receiver_coords_arr = num.empty((len(targets), 5))
            for itarget, target in enumerate(targets):
                receiver = target.receiver(store)
                receiver_coords_arr[itarget, :] = \
                    [receiver.lat, receiver.lon, receiver.north_shift,
                     receiver.east_shift, receiver.depth]

            delays = dsource.times

            itmin = num.zeros(ntargets, dtype=num.int32)
            nsamples = num.full(ntargets, -1, dtype=num.int32)

            if random_itmin:
                itmin = num.random.randint(-20, 5, size=ntargets,
                                           dtype=num.int32)
            if random_nsamples:
                nsamples = num.random.randint(10, 100, size=ntargets,
                                              dtype=num.int32)

            @benchmark.labeled('calc_timeseries-%s' % interpolation)
            def calc_timeseries():

                return store_ext.store_calc_timeseries(
                    store.cstore,
                    source_coords_arr,
                    source_terms,
                    delays,
                    receiver_coords_arr,
                    scheme,
                    interpolation,
                    itmin,
                    nsamples,
                    nthreads)

            @benchmark.labeled('sum_timeseries-%s' % interpolation)
            def sum_timeseries():
                results = []
                for itarget, target in enumerate(targets):
                    params = store_ext.make_sum_params(
                        store.cstore,
                        source_coords_arr,
                        source_terms,
                        target.coords5[num.newaxis, :].copy(),
                        scheme,
                        interpolation,
                        nthreads)

                    for weights, irecords in params:
                        neach = irecords.size // dsource.times.size
                        delays2 = num.repeat(dsource.times, neach)
                        r = store_ext.store_sum(
                            store.cstore,
                            irecords,
                            delays2,
                            weights,
                            int(itmin[itarget]),
                            int(nsamples[itarget]))

                        results.append(r)

                return results

            res_calc = calc_timeseries()
            res_sum = sum_timeseries()

            for c, s in zip(res_calc, res_sum):
                num.testing.assert_equal(c[0], s[0], verbose=True)
                for cc, cs in zip(c[1:-1], s[1:]):
                    assert cc == cs

        source_store_targets = [
            (
                gf.ExplosionSource(time=0.0, depth=100., moment=1.0),
                gf.Store(self.get_pulse_store_dir()),
                [
                    gf.Target(
                        codes=('', 'STA', '', component),
                        north_shift=500.,
                        east_shift=500.,
                        tmin=-300.,
                        tmax=+300.)
                    for component in 'ZNE'
                ]
            ),
            (
                gf.RectangularSource(
                    lat=0., lon=0., depth=5*km,
                    length=1*km, width=1*km, anchor='top'),
                engine.get_store('global_2s'),
                [
                    gf.Target(
                        codes=('', 'STA%02i' % istation, '', component),
                        lat=lat,
                        lon=lon,
                        store_id='global_2s')
                    for component in 'ZNE'
                    for istation, (lat, lon) in enumerate(
                        num.random.random((1, 2)))
                ]
            )
        ]

        for source, store, targets in source_store_targets:
            store.open()
            for interp in ['multilinear', 'nearest_neighbor']:
                for random_itmin in [True, False]:
                    for random_nsamples in [True, False]:
                        test_timeseries(
                            store, source, targets,
                            interpolation=interp,
                            nthreads=1,
                            random_itmin=random_itmin,
                            random_nsamples=random_nsamples)

                print(benchmark)
                benchmark.clear()

    @unittest.skipIf('global_2s' not in local_stores,
                     'depends on store global_2s')
    def test_process_timeseries(self):
        engine = gf.LocalEngine(use_config=True)

        sources = [
            gf.ExplosionSource(
                time=0.0,
                depth=depth,
                moment=moment)

            for moment in [2., 4., 8.] for depth in [3000., 6000., 12000.]
        ]

        targets = [
            gf.Target(
                codes=('', 'ST%d' % i, '', component),
                north_shift=shift*km,
                east_shift=0.,
                tmin=tmin,
                store_id='global_2s',
                tmax=None if tmin is None else tmin+40.)

            for component in 'ZNE' for i, shift in enumerate([100])
            for tmin in [None, 5., 20.]
        ]

        response_sum = engine.process(sources=sources, targets=targets,
                                      calc_timeseries=False, nthreads=1)

        response_calc = engine.process(sources=sources, targets=targets,
                                       calc_timeseries=True, nthreads=1)

        for (source, target, tr), (source_n, target_n, tr_n) in zip(
                response_sum.iter_results(), response_calc.iter_results()):
            assert source is source_n
            assert target is target_n

            t1 = tr.get_xdata()
            t2 = tr_n.get_xdata()
            num.testing.assert_equal(t1, t2)

            disp1 = tr.get_ydata()
            disp2 = tr_n.get_ydata()

            num.testing.assert_equal(disp1, disp2)

    def _test_homogeneous_scenario(
            self,
            config_type_class,
            component_scheme,
            discretized_source_class):

        if config_type_class.short_type == 'C' \
                or component_scheme.startswith('poro'):

            assert False

        store_id = 'homogeneous_%s_%s' % (
            config_type_class.short_type, component_scheme)

        vp = 5.8 * km
        vs = 3.46 * km

        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
  0. %(vp)g %(vs)g 2.6 1264. 600.
 20. %(vp)g %(vs)g 2.6 1264. 600.'''.lstrip() % dict(vp=vp/km, vs=vs/km)))

        store_type = config_type_class.short_type
        params = dict(
            id=store_id,
            sample_rate=1000.,
            modelling_code_id='ahfullgreen',
            component_scheme=component_scheme,
            earthmodel_1d=mod)

        if store_type in ('A', 'B'):
            params.update(
                source_depth_min=1.*km,
                source_depth_max=2.*km,
                source_depth_delta=0.5*km,
                distance_min=4.*km,
                distance_max=6.*km,
                distance_delta=0.5*km)

        if store_type == 'A':
            params.update(
                receiver_depth=3.*km)

        if store_type == 'B':
            params.update(
                receiver_depth_min=2.*km,
                receiver_depth_max=3.*km,
                receiver_depth_delta=0.5*km)

        if store_type == 'C':
            params.update(
                source_depth_min=1.*km,
                source_depth_max=2.*km,
                source_depth_delta=0.5*km,
                source_east_shift_min=1.*km,
                source_east_shift_max=2.*km,
                source_east_shift_delta=0.5*km,
                source_north_shift_min=2.*km,
                source_north_shift_max=3.*km,
                source_north_shift_delta=0.5*km)

        config = config_type_class(**params)

        config.validate()

        store_dir = mkdtemp(prefix=store_id)
        self.tempdirs.append(store_dir)

        gf.store.Store.create_editables(store_dir, config=config)

        store = gf.store.Store(store_dir, 'r')
        store.make_ttt()
        store.close()

        fomosto_ahfullgreen.build(store_dir, nworkers=1)

        store = gf.store.Store(store_dir, 'r')

        dsource_type = discretized_source_class.__name__

        params = {}
        if dsource_type == 'DiscretizedMTSource':
            params.update(
                m6s=num.array([
                    [1., 2., 3., 4., 5., 6.],
                    [1., 2., 3., 4., 5., 6.]]))
        elif dsource_type == 'DiscretizedExplosionSource':
            params.update(
                m0s=num.array([2., 2.]))
        elif dsource_type == 'DiscretizedSFSource':
            params.update(
                forces=num.array([[1., 2., 3.], [1., 2., 3.]]))
        elif dsource_type == 'DiscretizedPorePressureSource':
            params.update(
                pp=num.array([3., 3.]))

        snorth = 2.0*km
        seast = 2.0*km
        sdepth = 1.0*km
        rnorth = snorth + 3.*km
        reast = seast + 4.*km
        rdepth = 3.0*km

        t0 = 10.0 * store.config.deltat

        dsource = discretized_source_class(
            times=num.array([t0, t0+5.*store.config.deltat]),
            north_shifts=num.array([snorth, snorth]),
            east_shifts=num.array([seast, seast]),
            depths=num.array([sdepth, sdepth]),
            **params)

        receiver = gf.Receiver(
            north_shift=rnorth,
            east_shift=reast,
            depth=rdepth)

        components = gf.component_scheme_to_description[
            component_scheme].provided_components

        for seismogram in [store.seismogram, store.calc_seismograms]:
            for interpolation in ['nearest_neighbor', 'multilinear']:
                trs1 = []

                res = seismogram(dsource, receiver, components,
                                 interpolation=interpolation)
                if isinstance(res, list) and len(res) == 1:
                    res = res[0]
                for component, gtr in res.items():

                    tr = gtr.to_trace('', 'STA', seismogram.__name__,
                                      component)
                    trs1.append(tr)

                trs2 = _make_traces_homogeneous(
                    dsource, receiver,
                    store.config.earthmodel_1d.require_homogeneous(),
                    store.config.deltat, '', 'STA', 'reference')

                tmin = max(tr.tmin for tr in trs1+trs2)
                tmax = min(tr.tmax for tr in trs1+trs2)
                for tr in trs1+trs2:
                    tr.chop(tmin, tmax)
                    assert tr.data_len() > 2

                trs1.sort(key=lambda tr: tr.channel)
                trs2.sort(key=lambda tr: tr.channel)

                denom = 0.0
                for t1, t2 in zip(trs1, trs2):
                    assert t1.channel == t2.channel
                    denom += num.sum(t1.ydata**2) + num.sum(t2.ydata**2)

                ds = []
                for t1, t2 in zip(trs1, trs2):
                    ds.append(2.0 * num.sum((t1.ydata - t2.ydata)**2) / denom)

                ds = num.array(ds)

                if component_scheme == 'elastic8':
                    limit = 1e-2
                else:
                    limit = 1e-6

                if not num.all(ds < limit):
                    trace.snuffle(trs1+trs2)

                assert num.all(ds < limit)


for config_type_class in gf.config_type_classes:
    for scheme in config_type_class.provided_schemes:
        for discretized_source_class in gf.discretized_source_classes:
            if scheme in discretized_source_class.provided_schemes:
                name = 'test_homogeneous_scenario_%s_%s_%s' % (
                    config_type_class.short_type,
                    scheme,
                    discretized_source_class.__name__)

                def make_method(
                        config_type_class, scheme, discretized_source_class):

                    @unittest.skipIf(
                        scheme.startswith('poro')
                        or config_type_class.short_type == 'C',
                        'todo: test poro and store type C')
                    def test_homogeneous_scenario(self):
                        return self._test_homogeneous_scenario(
                            config_type_class,
                            scheme,
                            discretized_source_class)

                    test_homogeneous_scenario.__name__ = name

                    return test_homogeneous_scenario
                setattr(GFTestCase, name, make_method(
                    config_type_class, scheme, discretized_source_class))


if __name__ == '__main__':
    util.setup_logging('test_gf', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import os
import tempfile
import unittest
from . import common

from pyrocko import util, trace
from pyrocko.io import rdseed
from pyrocko.io import eventdata


class RDSeedTestCase(unittest.TestCase):

    def test_read(self):
        if not rdseed.Programs.check():
            raise unittest.SkipTest('rdseed executeable not found.')

        fpath = common.test_data_file('test_stations.dseed')
        dseed = rdseed.SeedVolumeAccess(fpath)
        pyrocko_stations = dseed.get_pyrocko_stations()
        assert(len(pyrocko_stations) == 1)
        assert(len(dseed.get_pyrocko_events()) == 0)

        tr = trace.Trace(tmax=1., network='CX', station='PB01', channel='HHZ')
        assert(isinstance(dseed.get_pyrocko_response(tr, target='dis'),
                          trace.Evalresp))

        st = dseed.get_pyrocko_station(tr)
        assert(st.network == tr.network)
        assert(st.network == tr.network)

    def test_problems(self):
        p = eventdata.Problems()
        p.add('X', 'y')
        fn = tempfile.mkstemp()[1]
        p.dump(fn)

        pl = eventdata.Problems()
        pl.load(fn)

        os.remove(fn)

        assert p.__dict__ == pl.__dict__


if __name__ == '__main__':
    util.setup_logging('test_rdseed', 'warning')
    unittest.main()



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
# python 2/3
from __future__ import absolute_import

import os
import math
import numpy as num
from builtins import str as newstr

from pyrocko import trace, util, plot
from pyrocko.guts import Object, Int, String, Timestamp

from . import io_common

N_GPS_TAGS_WANTED = 200  # must match definition in datacube_ext.c


def color(c):
    c = plot.color(c)
    return tuple(x/255. for x in c)


class DataCubeError(io_common.FileLoadError):
    pass


def make_control_point(ipos_block, t_block, tref, deltat):

    # reduce time (no drift would mean straight line)
    tred = (t_block - tref) - ipos_block*deltat

    # first round, remove outliers
    q25, q75 = num.percentile(tred, (25., 75.))
    iok = num.where(num.logical_and(q25 <= tred, tred <= q75))[0]

    # detrend
    slope, offset = num.polyfit(ipos_block[iok], tred[iok], 1)
    tred2 = tred - (offset + slope * ipos_block)

    # second round, remove outliers based on detrended tred, refit
    q25, q75 = num.percentile(tred2, (25., 75.))
    iok = num.where(num.logical_and(q25 <= tred2, tred2 <= q75))[0]
    slope, offset = num.polyfit(ipos_block[iok], tred[iok], 1)

    ic = ipos_block[ipos_block.size//2]
    tc = offset + slope * ic

    return ic, tc + ic * deltat + tref


def analyse_gps_tags(header, gps_tags, offset, nsamples):

    ipos, t, fix, nsvs = gps_tags

    n = t.size
    ok = num.logical_and(
        abs(t[1:n-1] - t[0:n-2] - 1.0) < 0.1,
        abs(t[2:n] - t[1:n-1] - 1.0) < 0.1)

    ipos = ipos[1:-1][ok]
    t = t[1:-1][ok]
    fix = fix[1:-1][ok]
    nsvs = nsvs[1:-1][ok]

    blocksize = N_GPS_TAGS_WANTED // 2
    deltat = 1.0 / int(header['S_RATE'])

    if ipos.size < blocksize:
        tmin = util.str_to_time(header['S_DATE'] + header['S_TIME'],
                                format='%y/%m/%d%H:%M:%S') + offset * deltat

        tmax = tmin + (nsamples - 1) * deltat
        icontrol, tcontrol = None, None
        return tmin, tmax, icontrol, tcontrol

    else:
        j = 0
        control_points = []
        tref = num.median(t - ipos*deltat)
        while j < ipos.size - blocksize:
            ipos_block = ipos[j:j+blocksize]
            t_block = t[j:j+blocksize]
            ic, tc = make_control_point(ipos_block, t_block, tref, deltat)
            control_points.append((ic, tc))
            j += blocksize

        ipos_last = ipos[-blocksize:]
        t_last = t[-blocksize:]
        ic, tc = make_control_point(ipos_last, t_last, tref, deltat)
        control_points.append((ic, tc))

        i0, t0 = control_points[0]
        i1, t1 = control_points[1]
        i2, t2 = control_points[-2]
        i3, t3 = control_points[-1]
        if len(control_points) == 2:
            tmin = t0 - i0 * deltat - offset * deltat
            tmax = t3 + (nsamples - i3 - 1) * deltat
        else:
            tmin = t0 + (offset - i0) * (t1 - t0) / (i1 - i0)
            tmax = t2 + (offset + nsamples - 1 - i2) * (t3 - t2) / (i3 - i2)

        if offset < i0:
            control_points[0:0] = [(offset, tmin)]

        if offset + nsamples - 1 > i3:
            control_points.append((offset + nsamples - 1, tmax))

        icontrol = num.array([x[0] for x in control_points], dtype=num.int64)
        tcontrol = num.array([x[1] for x in control_points], dtype=num.float)

        return tmin, tmax, icontrol, tcontrol


def plot_timeline(fns):
    from matplotlib import pyplot as plt

    fig = plt.figure()
    axes = fig.gca()

    h = 3600.

    if isinstance(fns, (str, newstr)):
        ipos, t, fix, nsvs, header, offset, nsamples = \
            get_extended_timing_context(fns)

    else:
        ipos, t, fix, nsvs, header, offset, nsamples = \
            get_timing_context(fns)

    deltat = 1.0 / int(header['S_RATE'])

    tref = num.median(t - ipos * deltat)
    tref = round(tref / deltat) * deltat

    x = ipos*deltat
    y = (t - tref) - ipos*deltat

    ifix = num.where(fix != 0)
    inofix = num.where(fix == 0)

    axes.plot(x[ifix]/h, y[ifix], '+', ms=5, color=color('chameleon3'))
    axes.plot(x[inofix]/h, y[inofix], 'x', ms=5, color=color('scarletred1'))

    tmin, tmax, icontrol, tcontrol = analyse_gps_tags(
        header, (ipos, t, fix, nsvs), offset, nsamples)

    tred = tcontrol - icontrol*deltat - tref
    axes.plot(icontrol*deltat/h, tred, color=color('aluminium6'))
    axes.plot(icontrol*deltat/h, tred, 'o', ms=5, color=color('aluminium6'))

    ymin = math.floor(tred.min() / deltat) * deltat - 0.1 * deltat
    ymax = math.ceil(tred.max() / deltat) * deltat + 0.1 * deltat

    # axes.set_ylim(ymin, ymax)
    if ymax - ymin < 100 * deltat:
        ygrid = math.floor(tred.min() / deltat) * deltat
        while ygrid < ymax:
            axes.axhline(ygrid, color=color('aluminium4'))
            ygrid += deltat

    xmin = icontrol[0]*deltat/h
    xmax = icontrol[-1]*deltat/h
    xsize = xmax - xmin
    xmin -= xsize * 0.1
    xmax += xsize * 0.1
    axes.set_xlim(xmin, xmax)

    axes.set_xlabel('Uncorrected (quartz) time [h]')
    axes.set_ylabel('Relative time correction [s]')

    plt.show()


g_dir_contexts = {}


class DirContextEntry(Object):
    path = String.T()
    tstart = Timestamp.T()
    ifile = Int.T()


class DirContext(Object):
    path = String.T()
    mtime = Timestamp.T()
    entries = DirContextEntry.T()

    def get_entry(self, fn):
        path = os.path.abspath(fn)
        for entry in self.entries:
            if entry.path == path:
                return entry

        raise Exception('entry not found')

    def iter_entries(self, fn, step=1):
        current = self.get_entry(fn)
        by_ifile = dict(
            (entry.ifile, entry) for entry in self.entries
            if entry.tstart == current.tstart)

        icurrent = current.ifile
        while True:
            icurrent += step
            try:
                yield by_ifile[icurrent]

            except KeyError:
                break


def context(fn):
    from pyrocko import datacube_ext

    dpath = os.path.dirname(os.path.abspath(fn))
    mtimes = [os.stat(dpath)[8]]

    dentries = sorted([os.path.join(dpath, f) for f in os.listdir(dpath)
                       if os.path.isfile(os.path.join(dpath, f))])
    for dentry in dentries:
        fn2 = os.path.join(dpath, dentry)
        mtimes.append(os.stat(fn2)[8])

    mtime = float(max(mtimes))

    if dpath in g_dir_contexts:
        dir_context = g_dir_contexts[dpath]
        if dir_context.mtime == mtime:
            return dir_context

        del g_dir_contexts[dpath]

    entries = []
    for dentry in dentries:
        fn2 = os.path.join(dpath, dentry)
        if not os.path.isfile(fn2):
            continue

        with open(fn2, 'rb') as f:
            first512 = f.read(512)
            if not detect(first512):
                continue

        with open(fn2, 'rb') as f:
            try:
                header, data_arrays, gps_tags, nsamples, _ = \
                        datacube_ext.load(f.fileno(), 3, 0, -1, None)

            except datacube_ext.DataCubeError as e:
                e = DataCubeError(str(e))
                e.set_context('filename', fn)
                raise e

        header = dict(header)
        entries.append(DirContextEntry(
            path=os.path.abspath(fn2),
            tstart=util.str_to_time(
                '20' + header['S_DATE'] + ' ' + header['S_TIME'],
                format='%Y/%m/%d %H:%M:%S'),
            ifile=int(header['DAT_NO'])))

    dir_context = DirContext(mtime=mtime, path=dpath, entries=entries)

    return dir_context


def get_time_infos(fn):
    from pyrocko import datacube_ext

    with open(fn, 'rb') as f:
        try:
            header, _, gps_tags, nsamples, _ = datacube_ext.load(
                f.fileno(), 1, 0, -1, None)

        except datacube_ext.DataCubeError as e:
            e = DataCubeError(str(e))
            e.set_context('filename', fn)
            raise e

    return dict(header), gps_tags, nsamples


def get_timing_context(fns):
    joined = [[], [], [], []]
    ioff = 0
    for fn in fns:
        header, gps_tags, nsamples = get_time_infos(fn)

        ipos = gps_tags[0]
        ipos += ioff

        for i in range(4):
            joined[i].append(gps_tags[i])

        ioff += nsamples

    ipos, t, fix, nsvs = [num.concatenate(x) for x in joined]

    nsamples = ioff
    return ipos, t, fix, nsvs, header, 0, nsamples


def get_extended_timing_context(fn):
    c = context(fn)

    header, gps_tags, nsamples_base = get_time_infos(fn)

    ioff = 0
    aggregated = [gps_tags]

    nsamples_total = nsamples_base

    if num.sum(gps_tags[2]) == 0:

        ioff = nsamples_base
        for entry in c.iter_entries(fn, 1):

            _, gps_tags, nsamples = get_time_infos(entry.path)

            ipos = gps_tags[0]
            ipos += ioff

            aggregated.append(gps_tags)
            nsamples_total += nsamples

            if num.sum(gps_tags[2]) > 0:
                break

            ioff += nsamples

        ioff = 0
        for entry in c.iter_entries(fn, -1):

            _, gps_tags, nsamples = get_time_infos(entry.path)

            ioff -= nsamples

            ipos = gps_tags[0]
            ipos += ioff

            aggregated[0:0] = [gps_tags]

            nsamples_total += nsamples

            if num.sum(gps_tags[2]) > 0:
                break

    ipos, t, fix, nsvs = [num.concatenate(x) for x in zip(*aggregated)]

#    return ipos, t, fix, nsvs, header, ioff, nsamples_total
    return ipos, t, fix, nsvs, header, 0, nsamples_base


def iload(fn, load_data=True, interpolation='sinc'):
    from pyrocko import datacube_ext
    from pyrocko import signal_ext

    if interpolation not in ('sinc', 'off'):
        raise NotImplemented(
            'no such interpolation method: %s' % interpolation)

    with open(fn, 'rb') as f:
        if load_data:
            loadflag = 2
        else:
            if interpolation == 'off':
                loadflag = 0
            else:
                # must get correct nsamples if interpolation is off
                loadflag = 1

        try:
            header, data_arrays, gps_tags, nsamples, _ = datacube_ext.load(
                f.fileno(), loadflag, 0, -1, None)

        except datacube_ext.DataCubeError as e:
            e = DataCubeError(str(e))
            e.set_context('filename', fn)
            raise e

    header = dict(header)
    deltat = 1.0 / int(header['S_RATE'])
    nchannels = int(header['CH_NUM'])

    ipos, t, fix, nsvs, header_, offset_, nsamples_ = \
        get_extended_timing_context(fn)

    tmin, tmax, icontrol, tcontrol = analyse_gps_tags(
        header_, (ipos, t, fix, nsvs), offset_, nsamples_)

    tmin_ip = round(tmin / deltat) * deltat
    if interpolation != 'off':
        tmax_ip = round(tmax / deltat) * deltat
    else:
        tmax_ip = tmin_ip + (nsamples-1) * deltat

    nsamples_ip = int(round((tmax_ip - tmin_ip)/deltat)) + 1
    # to prevent problems with rounding errors:
    tmax_ip = tmin_ip + (nsamples_ip-1) * deltat

    leaps = num.array(
        [x[0] + util.gps_utc_offset(x[0]) for x in util.read_leap_seconds2()],
        dtype=num.float)

    for i in range(nchannels):
        if load_data:
            arr = data_arrays[i]
            assert arr.size == nsamples

            if interpolation == 'sinc' and icontrol is not None:
                ydata = num.empty(nsamples_ip, dtype=num.float)
                signal_ext.antidrift(
                    icontrol, tcontrol,
                    arr.astype(num.float), tmin_ip, deltat, ydata)

                ydata = num.round(ydata).astype(arr.dtype)
            else:
                ydata = arr

            tr_tmin = tmin_ip
            tr_tmax = None
        else:
            ydata = None
            tr_tmin = tmin_ip
            tr_tmax = tmax_ip

        tr = trace.Trace('', header['DEV_NO'], '', 'p%i' % i, deltat=deltat,
                         ydata=ydata, tmin=tr_tmin, tmax=tr_tmax, meta=header)

        bleaps = num.logical_and(tmin_ip <= leaps, leaps < tmax_ip)

        if num.any(bleaps):
            assert num.sum(bleaps) == 1
            tcut = leaps[bleaps][0]

            for tmin_cut, tmax_cut in [
                    (tr.tmin, tcut), (tcut, tr.tmax+tr.deltat)]:

                try:
                    tr_cut = tr.chop(tmin_cut, tmax_cut, inplace=False)
                    tr_cut.shift(
                        util.utc_gps_offset(0.5*(tr_cut.tmin+tr_cut.tmax)))
                    yield tr_cut

                except trace.NoData:
                    pass

        else:
            tr.shift(util.utc_gps_offset(0.5*(tr.tmin+tr.tmax)))
            yield tr


header_keys = {
    str: b'GIPP_V DEV_NO E_NAME GPS_PO S_TIME S_DATE DAT_NO'.split(),
    int: b'''P_AMPL CH_NUM S_RATE D_FILT C_MODE A_CHOP F_TIME GPS_TI GPS_OF
            A_FILT A_PHAS GPS_ON ACQ_ON V_TCXO D_VOLT E_VOLT'''.split()}

all_header_keys = header_keys[str] + header_keys[int]


def detect(first512):
    s = first512

    if len(s) < 512:
        return False

    if ord(s[0:1]) >> 4 != 15:
        return False

    n = s.find(b'\x80')
    if n == -1:
        n = len(s)

    s = s[1:n]
    s = s.replace(b'\xf0', b'')
    s = s.replace(b';', b' ')
    s = s.replace(b'=', b' ')
    kvs = s.split(b' ')

    if len([k for k in all_header_keys if k in kvs]) == 0:
        return False
    return True


if __name__ == '__main__':
    import sys
    fns = sys.argv[1:]
    if len(fns) > 1:
        plot_timeline(fns)
    else:
        plot_timeline(fns[0])



# python 2/3

from __future__ import division, print_function, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa
from builtins import range
import unittest
import math
import numpy as num
from io import BytesIO

from pyrocko import util, moment_tensor as mtm
from pyrocko.plot import beachball

from random import random, choice


def fuzz_angle(mi, ma):
    crits = [mi, ma, 0., 1., -1., 90., -90., 180., -180., 270., -270.]
    crits = [x for x in crits if mi <= x and x <= ma]
    ran = [
        0.,
        0.,
        0.,
        +random()*(ma-mi)*1.0e-8,
        -random()*(ma-mi)*1.0e-8,
        +random()*(ma-mi)*1.0e-4,
        -random()*(ma-mi)*1.0e-4,
        random()*(ma-mi),
        -random()*(ma-mi)]

    while True:
        v = choice(crits) + choice(ran)
        if mi <= v and v <= ma:
            return v


class BeachballTestCase(unittest.TestCase):

    def compare_beachball(self, mt, show=False, **kwargs):
        from matplotlib import pyplot as plt
        from matplotlib import image
        plt.switch_backend('Agg')

        plotargs = dict(
            size=1.0,
            size_units='data',
            linewidth=2.0,
            projection='lambert')

        plotargs.update(kwargs)

        imgs = []
        for iplot, plot in enumerate([
                beachball.plot_beachball_mpl,
                beachball.plot_beachball_mpl_pixmap]):

            fig = plt.figure(figsize=(3, 3), dpi=100)
            axes = fig.add_subplot(1, 1, 1, aspect=1.)
            axes.cla()
            axes.axison = False
            axes.set_xlim(-1.05, 1.05)
            axes.set_ylim(-1.05, 1.05)

            plot(mt, axes, **plotargs)

            f = BytesIO()
            fig.savefig(f, format='png')
            f.seek(0)
            imgs.append(image.imread(f, format='png'))
            fig.clear()
            plt.close(fig)

        a, b = imgs

        d = num.abs(a-b)
        d[:, :, 3] = 1.
        dsum = num.sum(d[:, :, :3])
        if dsum > 1600 or show:
            print(dsum)
            print(mt)
            fig = plt.figure()
            axes1 = fig.add_subplot(1, 3, 1, aspect=1.)
            axes2 = fig.add_subplot(1, 3, 2, aspect=1.)
            axes3 = fig.add_subplot(1, 3, 3, aspect=1.)
            axes1.imshow(a)
            axes2.imshow(b)
            axes3.imshow(d)
            plt.show()
            plt.close(fig)

        assert dsum <= 1600
        return dsum

    def test_random_mts(self, **kwargs):
        nx = 10
        for x in range(nx):
            m6 = num.random.random(6)*2.-1.
            m = mtm.symmat6(*m6)
            mt = mtm.MomentTensor(m=m)
            self.compare_beachball(mt, **kwargs)

        for x in range(nx):
            mt = mtm.MomentTensor.random_mt()
            self.compare_beachball(mt, **kwargs)

    def test_projections(self):
        n = 1000
        points = num.zeros((n, 3))
        phi = num.random.random(n) * math.pi * 2.0
        theta = num.random.random(n) * math.pi / 2.0

        points[:, 0] = num.cos(phi) * num.sin(theta)
        points[:, 1] = num.sin(phi) * num.sin(theta)
        points[:, 2] = num.cos(theta)

        for projection in ['lambert', 'stereographic', 'orthographic']:
            projection = 'stereographic'
            points2 = beachball.project(points, projection=projection)
            points_new = beachball.inverse_project(
                points2, projection=projection)
            assert num.all(num.abs(points - points_new) < 1e-5)

    def test_specific_mts(self):
        for m6 in [
                (1., 0., 0., 0., 0., 0.),
                (0., 1., 0., 0., 0., 0.),
                (0., 0., 1., 0., 0., 0.),
                (0., 0., 0., 1., 0., 0.),
                (0., 0., 0., 0., 1., 0.),
                (0., 0., 0., 0., 0., 1.)]:

            m = mtm.symmat6(*m6)
            mt = mtm.MomentTensor(m=m)

            self.compare_beachball(mt)

    def test_random_dcs(self):
        nx = 20

        for x in range(nx):
            strike = fuzz_angle(0., 360.)
            dip = fuzz_angle(0., 90.)
            rake = fuzz_angle(-180., 180.)

            mt = mtm.MomentTensor(
                strike=strike,
                dip=dip,
                rake=rake)

            self.compare_beachball(mt)

    def test_specific_dcs(self):
        for strike, dip, rake in [
                [270., 0.0, 0.01],
                [360., 28.373841741182012, 90.],
                [0., 0., 0.]]:

            mt = mtm.MomentTensor(
                strike=strike,
                dip=dip,
                rake=rake)

            self.compare_beachball(mt)

    def test_random_mts_views(self):
        for view in ('top', 'north', 'south', 'east', 'west'):
            self.test_random_mts(view=view)

    @unittest.skip('contour and contourf do not support transform')
    def test_plotstyle(self):

        # contour and contourf do not support transform
        mt = mtm.MomentTensor.random_mt()
        self.compare_beachball(
            mt, show=True, size=20., size_units='points', position=(0.5, 0.5))

    def test_fuzzy_beachball(self):

        from matplotlib import pyplot as plt

        def get_random_uniform(lower, upper, dimension=1):
            return (upper - lower) * num.random.rand(dimension) + lower

        fig = plt.figure(figsize=(4., 4.))
        fig.subplots_adjust(left=0., right=1., bottom=0., top=1.)
        axes = fig.add_subplot(1, 1, 1)

        strike = 135.
        dip = 65.
        rake = 15.

        best_mt = mtm.MomentTensor.from_values((strike, dip, rake))

        n_balls = 1000
        mts = []
        for i in range(n_balls):
            strike_dev = get_random_uniform(-15., 15.)
            mts.append(mtm.MomentTensor.from_values(
                (strike + strike_dev, dip, rake)))

        kwargs = {
            'beachball_type': 'full',
            'size': 8,
            'position': (5, 5),
            'color_t': 'black',
            'edgecolor': 'black'}

        beachball.plot_fuzzy_beachball_mpl_pixmap(mts, axes, best_mt, **kwargs)

        axes.set_xlim(0., 10.)
        axes.set_ylim(0., 10.)
        axes.set_axis_off()

        # fig.savefig('fuzzy_bb_no.png', dpi=400)
        # plt.show()

    def show_comparison(self, nx=10):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D  # noqa

        if nx > 1:
            plt.ion()
            plt.show()

        fig = plt.figure()
        axes1 = fig.add_subplot(2, 3, 1, aspect=1.)
        axes2 = fig.add_subplot(2, 3, 2, aspect=1.)
        axes3 = fig.add_subplot(2, 3, 3, aspect=1.)
        axes4 = fig.add_subplot(2, 3, 4, projection='3d', aspect=1.)
        axes5 = fig.add_subplot(2, 3, 5, projection='3d', aspect=1.)

        try:
            import mopad
        except ImportError:
            mopad = None

        for x in range(nx):
            mt = mtm.MomentTensor.random_mt()
            mt = mt.deviatoric()

            for axes in (axes1, axes2, axes3):
                axes.cla()
                axes.axison = False
                axes.set_xlim(-1.05, 1.05)
                axes.set_ylim(-1.05, 1.05)

            for axes in (axes4, axes5):
                axes.cla()

            axes1.set_title('Copacabana')
            axes2.set_title('Contour')
            axes3.set_title('MoPaD')
            axes4.set_title('Patches')
            axes5.set_title('Lines')

            beachball.plot_beachball_mpl(mt, axes1, size_units='data')
            beachball.plot_beachball_mpl_pixmap(mt, axes2, size_units='data')

            beachball.plot_beachball_mpl_construction(
                mt, axes4, show='patches')
            beachball.plot_beachball_mpl_construction(
                mt, axes5, show='lines')

            if mopad:
                try:
                    mop_mt = mopad.MomentTensor(M=mt.m6())
                    mop_beach = mopad.BeachBall(mop_mt)
                    kwargs = dict(
                        plot_projection='lambert',
                        plot_nodalline_width=2,
                        plot_faultplane_width=2,
                        plot_outerline_width=2)

                    mop_beach.ploBB(kwargs, ax=axes3)

                except Exception:
                    print(
                        'mopad failed (maybe patched mopad version is needed')

            fig.canvas.draw()

        if nx == 1:
            plt.show()


if __name__ == "__main__":
    util.setup_logging('test_beachball', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa

import unittest
import tempfile
import numpy as num
import urllib
import logging
from pyrocko import util, trace
from pyrocko.io import stationxml
from pyrocko.client import fdsn, iris

from . import common

logger = logging.getLogger('pyrocko.test.test_fdsn')

stt = util.str_to_time


class FDSNStationTestCase(unittest.TestCase):

    def test_read_samples(self):
        ok = False
        for fn in ['geeil.iris.xml', 'geeil.geofon.xml']:
            fpath = common.test_data_file(fn)
            x = stationxml.load_xml(filename=fpath)
            for network in x.network_list:
                assert network.code == 'GE'
                for station in network.station_list:
                    assert station.code == 'EIL'
                    for channel in station.channel_list:
                        assert channel.code[:2] == 'BH'
                        for stage in channel.response.stage_list:
                            ok = True

            assert ok

            pstations = x.get_pyrocko_stations()
            assert len(pstations) in (3, 4)
            for s in x.get_pyrocko_stations():
                assert len(s.get_channels()) == 3

            assert len(x.get_pyrocko_stations(
                time=stt('2010-01-15 10:00:00'))) == 1

            new = stationxml.FDSNStationXML.from_pyrocko_stations(pstations)
            assert len(new.get_pyrocko_stations()) in (3, 4)
            for s in new.get_pyrocko_stations():
                assert len(s.get_channels()) == 3

    @common.require_internet
    @common.skip_on_download_error
    def test_retrieve(self):
        for site in ['geofon', 'iris']:
            fsx = fdsn.station(site=site,
                               network='GE',
                               station='EIL',
                               level='channel')

            assert len(fsx.get_pyrocko_stations(
                time=stt('2010-01-15 10:00:00'))) == 1

    @common.require_internet
    @common.skip_on_download_error
    def test_dataselect(self):
        tmin = stt('2010-01-15 10:00:00')
        tmax = stt('2010-01-15 10:01:00')
        for site in ['geofon', 'iris']:
            fdsn.dataselect(site=site,
                            network='GE',
                            station='EIL',
                            starttime=tmin,
                            endtime=tmax)

    @common.require_internet
    @common.skip_on_download_error
    def test_dataselection(self):
        tmin = stt('2010-01-15 10:00:00')
        tmax = stt('2010-01-15 10:01:00')
        selection = [
            ('GE', 'EIL', '*', 'SHZ', tmin, tmax),
        ]

        fdsn.dataselect(site='geofon', selection=selection)
        fdsn.station(site='geofon', selection=selection, level='response')

    def test_read_big(self):
        for site in ['iris']:
            fpath = common.test_data_file('%s_1014-01-01_all.xml' % site)
            stationxml.load_xml(filename=fpath)

    # @unittest.skip('needs manual inspection')
    @common.require_internet
    @common.skip_on_download_error
    def test_response(self, ntest=4):
        tmin = stt('2014-01-01 00:00:00')
        tmax = stt('2014-01-02 00:00:00')
        sx = fdsn.station(
            site='iris',
            network='II',
            channel='?HZ',
            startbefore=tmin,
            endafter=tmax,
            level='channel', format='text', matchtimeseries=True)

        for nslc in sx.nslc_code_list[:ntest]:
            net, sta, loc, cha = nslc
            sxr = fdsn.station(
                site='iris',
                network=net,
                station=sta,
                location=loc,
                channel=cha,
                startbefore=tmin,
                endafter=tmax,
                level='response', matchtimeseries=True)

            fi = iris.ws_resp(
                network=net,
                station=sta,
                location=loc,
                channel=cha,
                tmin=tmin,
                tmax=tmax)

            _, fn = tempfile.mkstemp()
            fo = open(fn, 'wb')
            while True:
                d = fi.read(1024)
                if not d:
                    break

                fo.write(d)

            fo.close()

            resp_sx = sxr.get_pyrocko_response(
                nslc, timespan=(tmin, tmax),
                fake_input_units='M/S')

            resp_er = trace.Evalresp(fn, target='vel', nslc_id=nslc, time=tmin)
            fmin = 0.001
            fmax = 100.

            for _, _, channel in sxr.iter_network_station_channels(
                    net, sta, loc, cha, timespan=(tmin, tmax)):
                if channel.response:
                    fmax = channel.sample_rate.value * 0.5

            f = num.exp(num.linspace(num.log(fmin), num.log(fmax), 500))

            t_sx = resp_sx.evaluate(f)
            t_er = resp_er.evaluate(f)
            import pylab as lab

            abs_dif = num.abs(num.abs(t_sx) - num.abs(t_er)) / num.max(
                num.abs(t_er))

            mda = num.mean(abs_dif[f < 0.4*fmax])

            pha_dif = num.abs(num.angle(t_sx) - num.angle(t_er))

            mdp = num.mean(pha_dif[f < 0.4*fmax])

            if mda > 0.05 or mdp > 0.04:

                lab.gcf().add_subplot(2, 1, 1)
                lab.plot(f, num.abs(t_sx), color='black')
                lab.plot(f, num.abs(t_er), color='red')
                lab.axvline(fmax/2., color='black')
                lab.axvline(fmax, color='gray')
                lab.xscale('log')
                lab.yscale('log')

                lab.gcf().add_subplot(2, 1, 2)
                lab.plot(f, num.angle(t_sx), color='black')
                lab.plot(f, num.angle(t_er), color='red')
                lab.xscale('log')
                lab.show()

                assert False, \
                    'evalresp and stationxml responses differ: %s' % str(nslc)


if __name__ == '__main__':
    util.setup_logging('test_fdsn', 'warning')
    unittest.main()



from __future__ import absolute_import
import unittest
import tempfile
import shutil
import os
import numpy as num
from matplotlib import image, pyplot as plt

from pyrocko import util
from pyrocko.plot import response

from . import common

noshow = True


class ResponsePlotTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def fpath(self, fn):
        return os.path.join(self.tempdir, fn)

    def fpath_ref(self, fn):
        try:
            return common.test_data_file(fn)
        except util.DownloadError:
            return common.test_data_file_no_download(fn)

    def compare_with_ref(self, fname, tolerance, show=False):
        fpath = self.fpath(fname)
        fpath_ref = self.fpath_ref(fname)

        if not os.path.exists(fpath_ref):
            shutil.copy(fpath, fpath_ref)

        img = image.imread(fpath)
        img_ref = image.imread(fpath_ref)
        self.assertEqual(img.shape, img_ref.shape)
        d = num.abs(img - img_ref)
        merr = num.mean(d)
        if (merr > tolerance or show) and not noshow:
            fig = plt.figure()
            axes1 = fig.add_subplot(1, 3, 1, aspect=1.)
            axes2 = fig.add_subplot(1, 3, 2, aspect=1.)
            axes3 = fig.add_subplot(1, 3, 3, aspect=1.)
            axes1.imshow(img)
            axes1.set_title('Candidate')
            axes2.imshow(img_ref)
            axes2.set_title('Reference')
            axes3.imshow(d)
            axes3.set_title('Mean abs difference: %g' % merr)
            plt.show()
            plt.close(fig)

        assert merr <= tolerance

    def test_response_plot(self):
        for fn, format in [
                ('test1.resp', 'resp'),
                ('test1.sacpz', 'sacpz')]:

            fpath_resp = common.test_data_file(fn)
            fname = 'test_response_plot_%s.png' % fn
            fpath_png = self.fpath(fname)
            resps, labels = response.load_response_information(
                fpath_resp, format)
            labels = [lab[len(fpath_resp)+1:] or 'dummy' for lab in labels]

            response.plot(
                responses=resps, labels=labels, filename=fpath_png, dpi=50)
            # self.compare_with_ref(fname, 0.01)


if __name__ == "__main__":
    util.setup_logging('test_response_plot', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import

import unittest
import numpy as num
from pyrocko import util, evalresp, pz, trace, guts

from . import common


def plot_tfs(freqs, tfs):
    colors = ['black', 'red', 'blue']
    import pylab as lab
    lab.gcf().add_subplot(2, 1, 1)
    for itf, tf in enumerate(tfs):
        lab.plot(freqs, num.abs(tf), color=colors[itf])

    lab.xscale('log')
    lab.yscale('log')

    lab.gcf().add_subplot(2, 1, 2)
    for itf, tf in enumerate(tfs):
        lab.plot(freqs, num.angle(tf), color=colors[itf])

    lab.xscale('log')
    lab.show()


class ResponseTestCase(unittest.TestCase):

    def test_evalresp(self, plot=False):

        resp_fpath = common.test_data_file('test2.resp')

        freqs = num.logspace(num.log10(0.001), num.log10(10.), num=1000)

        transfer = evalresp.evalresp(
            sta_list='BSEG',
            cha_list='BHZ',
            net_code='GR',
            locid='',
            instant=util.str_to_time('2012-01-01 00:00:00'),
            freqs=freqs,
            units='DIS',
            file=resp_fpath,
            rtype='CS')[0][4]

        pz_fpath = common.test_data_file('test2.sacpz')

        zeros, poles, constant = pz.read_sac_zpk(pz_fpath)

        resp = trace.PoleZeroResponse(zeros, poles, constant)

        transfer2 = resp.evaluate(freqs)

        if plot:
            plot_tfs(freqs, [transfer, transfer2])

        assert numeq(transfer, transfer2, 1e-4)

    def test_conversions(self):

        from pyrocko import model
        from pyrocko.io import resp, enhanced_sacpz
        from pyrocko.io import stationxml

        t = util.str_to_time('2014-01-01 00:00:00')
        codes = 'GE', 'EIL', '', 'BHZ'

        resp_fpath = common.test_data_file('test1.resp')
        stations = [model.Station(
            *codes[:3],
            lat=29.669901,
            lon=34.951199,
            elevation=210.0,
            depth=0.0)]

        sx_resp = resp.make_stationxml(
            stations, resp.iload_filename(resp_fpath))

        pr_sx_resp = sx_resp.get_pyrocko_response(
            codes, time=t, fake_input_units='M/S')
        pr_evresp = trace.Evalresp(
            resp_fpath, nslc_id=codes, target='vel', time=t)

        sacpz_fpath = common.test_data_file('test1.sacpz')
        sx_sacpz = enhanced_sacpz.make_stationxml(
            enhanced_sacpz.iload_filename(sacpz_fpath))
        pr_sx_sacpz = sx_sacpz.get_pyrocko_response(
            codes, time=t, fake_input_units='M/S')
        pr_sacpz = trace.PoleZeroResponse(*pz.read_sac_zpk(sacpz_fpath))
        try:
            pr_sacpz.zeros.remove(0.0j)
        except ValueError:
            pr_sacpz.poles.append(0.0j)

        sxml_geofon_fpath = common.test_data_file('test1.stationxml')
        sx_geofon = stationxml.load_xml(filename=sxml_geofon_fpath)
        pr_sx_geofon = sx_geofon.get_pyrocko_response(
            codes, time=t, fake_input_units='M/S')

        sxml_iris_fpath = common.test_data_file('test2.stationxml')
        sx_iris = stationxml.load_xml(filename=sxml_iris_fpath)
        pr_sx_iris = sx_iris.get_pyrocko_response(
            codes, time=t, fake_input_units='M/S')

        freqs = num.logspace(num.log10(0.001), num.log10(1.0), num=1000)
        tf_ref = pr_evresp.evaluate(freqs)
        for pr in [pr_sx_resp, pr_sx_sacpz, pr_sacpz, pr_sx_geofon,
                   pr_sx_iris]:
            tf = pr.evaluate(freqs)
            # plot_tfs(freqs, [tf_ref, tf])
            assert cnumeqrel(tf_ref, tf, 0.01)

    def test_dump_load(self):

        r = trace.FrequencyResponse()

        r = trace.PoleZeroResponse([0j, 0j], [1j, 2j, 1+3j, 1-3j], 1.0)
        r.regularize()
        r2 = guts.load_string(r.dump())
        assert cnumeq(r.poles, r2.poles, 1e-6)
        assert cnumeq(r.zeros, r2.zeros, 1e-6)
        assert numeq(r.constant, r2.constant)

        r = trace.SampledResponse(
            [0., 1., 5., 10.],
            [0., 1., 1., 0.])

        r.regularize()
        r2 = guts.load_string(r.dump())
        assert numeq(r.frequencies, r2.frequencies, 1e-6)
        assert cnumeq(r.values, r2.values, 1e-6)

        r = trace.IntegrationResponse(2, 5.0)
        r2 = guts.load_string(r.dump())
        assert numeq(r.n, r2.n)
        assert numeq(r.gain, r2.gain, 1e-6)

        r = trace.DifferentiationResponse(2, 5.0)
        r2 = guts.load_string(r.dump())
        assert numeq(r.n, r2.n)
        assert numeq(r.gain, r2.gain, 1e-6)

        r = trace.AnalogFilterResponse(
            a=[1.0, 2.0, 3.0],
            b=[2.0, 3.0])
        r2 = guts.load_string(r.dump())
        assert numeq(r.a, r2.a, 1e-6)
        assert numeq(r.b, r2.b, 1e-6)


def numeq(a, b, eps=0.0):
    a = num.asarray(a)
    b = num.asarray(b)
    return num.all(num.abs(a - b) <= eps)


def cnumeq(a, b, eps=0.0):
    a = num.asarray(a)
    b = num.asarray(b)
    return num.all(num.abs(a.real - b.real) <= eps) and \
        num.max(num.abs(a.imag - b.imag) <= eps)


def cnumeqrel(a, b, eps=0.0):
    a = num.asarray(a)
    b = num.asarray(b)
    return num.all(num.abs(a - b)/(num.abs(a) + num.abs(b)) <= eps)


if __name__ == "__main__":
    util.setup_logging('test_response', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from pyrocko import util
from pyrocko.client import catalog
from pyrocko import moment_tensor
import unittest
from . import common


def near(a, b, eps):
    return abs(a-b) < eps


def is_the_haiti_event_geofon(ev):
    assert near(ev.magnitude, 7.2, 0.001)
    assert near(ev.lat, 18.37, 0.001)
    assert near(ev.lon, -72.55, 0.001)
    assert near(ev.depth, 17000., 1.)
    assert ev.region == 'Haiti Region'
    assert ev.tags.index('geofon_category:xxl') != -1


class CatalogTestCase(unittest.TestCase):

    @common.require_internet
    def testGeofon(self):

        cat = catalog.Geofon()

        tmin = util.str_to_time('2010-01-12 21:50:00')
        tmax = util.str_to_time('2010-01-13 03:17:00')

        names = cat.get_event_names(
            time_range=(tmin, tmax), nmax=10, magmin=5.)

        assert len(names) > 0
        ident = None
        for name in names:
            ev = cat.get_event(name)
            if ev.magnitude >= 7:
                is_the_haiti_event_geofon(ev)
                ident = ev.name

        assert ident is not None

        cat.flush()

    @common.require_internet
    def testGeofonSingleEvent(self):
        cat = catalog.Geofon()
        ev = cat.get_event('gfz2010avtm')
        is_the_haiti_event_geofon(ev)
        cat.flush()

        ev = cat.get_event('gfz2019hkck')
        assert ev.name == 'gfz2019hkck'

    @common.require_internet
    def testGeofonMT(self):
        cat = catalog.Geofon()
        tmin = util.str_to_time('2014-01-01 00:00:00')
        tmax = util.str_to_time('2017-01-01 00:00:00')
        events = cat.get_events((tmin, tmax), magmin=8)
        self.assertEqual(len(events), 2)
        mt1, mt2 = [ev.moment_tensor for ev in events]
        angle = moment_tensor.kagan_angle(mt1, mt2)
        self.assertEqual(round(angle - 7.7, 1), 0.0)

    @unittest.skip("don't want to annoy our colleagues at GEOFON")
    def testGeofonMassive(self):
        cat = catalog.Geofon(get_moment_tensors=False)
        tmin = util.str_to_time('2010-01-01 00:00:00')
        tmax = util.str_to_time('2018-01-01 00:00:00')
        events = cat.get_events((tmin, tmax))
        print(len(events))

    @common.require_internet
    def testISC(self):
        cat = catalog.ISC()
        tmin = util.stt('2009-02-22 05:00:00')
        tmax = util.stt('2009-02-22 19:00:00')
        cat.get_events((tmin, tmax), magmin=5)

        markers = cat.get_phase_markers(
            (tmin, tmax), phases=['XXX'], station_codes=['XXXXX', 'YYYYY'])

        self.assertEqual(len(markers), 0)

        markers = cat.get_phase_markers(
            (tmin, tmax), phases=['P', 'PcP'], station_codes=['WRA'])

        self.assertEqual(len(markers), 75)

    @common.require_internet
    def testGlobalCMT(self):

        def is_the_haiti_event(ev):
            assert near(ev.magnitude, 7.0, 0.1)
            assert near(ev.lat, 18.61, 0.01)
            assert near(ev.lon, -72.62, 0.01)
            assert near(ev.depth, 12000., 1.)
            assert ev.region.lower() == 'haiti region'

        cat = catalog.GlobalCMT()

        tmin = util.str_to_time('2010-01-12 21:50:00')
        tmax = util.str_to_time('2010-01-13 03:17:00')

        names = cat.get_event_names(time_range=(tmin, tmax), magmin=5.)
        ident = None
        for name in names:
            ev = cat.get_event(name)
            if ev.magnitude > 7:
                is_the_haiti_event(ev)
                ident = ev.name

        assert ident is not None
        cat.flush()
        ev = cat.get_event(ident)
        is_the_haiti_event(ev)

    @common.require_internet
    def testUSGS(self):

        def is_the_haiti_event(ev):
            assert near(ev.magnitude, 7.0, 0.1)
            assert near(ev.lat, 18.443, 0.01)
            assert near(ev.lon, -72.571, 0.01)
            assert near(ev.depth, 13000., 1.)

        cat = catalog.USGS()

        tmin = util.str_to_time('2010-01-12 21:50:00')
        tmax = util.str_to_time('2010-01-13 03:17:00')

        names = cat.get_event_names(time_range=(tmin, tmax), magmin=5.)
        assert len(names) == 13
        for name in names:
            ev = cat.get_event(name)
            if ev.magnitude >= 7.:
                is_the_haiti_event(ev)
                ident = ev.name

        assert ident is not None
        cat.flush()
        ev = cat.get_event(ident)
        is_the_haiti_event(ev)


if __name__ == "__main__":
    util.setup_logging('test_catalog', 'debug')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import str
from builtins import range
import unittest
import random
import math

import numpy as num

from pyrocko.moment_tensor import \
    magnitude_to_moment, moment_to_magnitude, MomentTensor, r2d, symmat6, \
    dynecm, kagan_angle, rotation_from_angle_and_axis, random_axis

from pyrocko import util, guts


class MomentTensorTestCase(unittest.TestCase):

    def testMagnitudeMoment(self):
        for i in range(1, 10):
            mag = float(i)
            assert abs(mag - moment_to_magnitude(magnitude_to_moment(mag))) \
                < 1e-6, 'Magnitude to moment to magnitude test failed.'

    def testAnyAngles(self):
        for i in range(100):
            (s1, d1, r1) = [r2d*random.random()*10.-5. for j in range(3)]
            m0 = 1.+random.random()*1.0e20
            self.forwardBackward(s1, d1, r1, m0)

    def testProblematicAngles(self):
        for i in range(100):
            # angles close to fractions of pi are especially problematic
            (s1, d1, r1) = [
                r2d*random.randint(-16, 16)*math.pi/8. +
                1e-8*random.random()-0.5e-8 for j in range(3)]
            m0 = 1.+random.random()*1.0e20
            self.forwardBackward(s1, d1, r1, m0)

    def testNonPlainDoubleCouples(self):
        for i in range(100):
            ms = [random.random()*1.0e20-0.5e20 for j in range(6)]
            m = num.matrix(
                [[ms[0], ms[3], ms[4]],
                 [ms[3], ms[1], ms[5]],
                 [ms[4], ms[5], ms[2]]], dtype=num.float)

            m1 = MomentTensor(m=m)
            m_plain = m1.m_plain_double_couple()
            m2 = MomentTensor(m=m_plain)

            self.assertAnglesSame(m1, m2)

    def forwardBackward(self, strike, dip, rake, scalar_moment):
        m1 = MomentTensor(
            strike=strike, dip=dip, rake=rake, scalar_moment=scalar_moment)
        m2 = MomentTensor(m=m1.m())
        self.assertAnglesSame(m1, m2)

    def assertAnglesSame(self, m1, m2):
        assert num.all(num.abs(num.array(m1.both_strike_dip_rake()) -
                               num.array(m2.both_strike_dip_rake()))
                       < 1e-7*100), \
            "angles don't match after forward-backward calculation:\n' \
            'first:\n"+str(m1) + "\nsecond:\n"+str(m2)

    def assertSame(self, a, b, eps, errstr):
        assert num.all(num.abs(num.array(a)-num.array(b)) < eps), errstr

    def testChile(self):
        m_use = symmat6(
            1.040, -0.030, -1.010, 0.227, -1.510, -0.120)*1e29*dynecm
        mt = MomentTensor(m_up_south_east=m_use)
        sdr = mt.both_strike_dip_rake()
        self.assertSame(sdr[0], (174., 73., 83.), 1., 'chile fail 1')
        self.assertSame(sdr[1], (18., 18., 112.), 1., 'chile fail 2')

    def testIO(self):
        m1 = MomentTensor(dip=90.)
        sdr1 = m1.both_strike_dip_rake()
        m2 = guts.load(string=m1.dump())
        sdr2 = m2.both_strike_dip_rake()
        self.assertSame(sdr1, sdr2, 0.1, 'failed io via guts')

    def testProps(self):
        m = MomentTensor()
        m.mnn = 1.
        m.mee = -1.
        m.mdd = 0.
        m.mne = 0.
        m.mnd = 0.
        m.med = 0.
        (s1, d1, _), (s2, d2, _) = m.both_strike_dip_rake()
        assert abs(s1 - 45.) < 0.1 or abs(s2 - 45.) < 0.1

    def testMomentGetterSetter(self):
        m1 = MomentTensor()
        want_mom = 2E7
        m1.moment = want_mom
        sm2 = m1.scalar_moment()
        assert(sm2 == m1.moment)
        assert(abs(sm2 - want_mom) < 1E-8)

        mag = moment_to_magnitude(want_mom)
        assert(m1.magnitude == mag)

        want_mag = 3.5
        m1.magnitude = want_mag
        mom = magnitude_to_moment(want_mag)
        assert(m1.moment == mom)

    def testKagan(self):
        eps = 0.01
        for _ in range(500):
            mt1 = MomentTensor.random_mt(magnitude=-1.0)
            assert 0.0 == kagan_angle(mt1, mt1)

            mt1 = MomentTensor.random_dc(magnitude=-1.0)
            assert 0.0 == kagan_angle(mt1, mt1)

            angle = random.random() * 90.

            rot = rotation_from_angle_and_axis(angle, random_axis())

            mrot = rot.T * mt1.m() * rot
            mt2 = MomentTensor(m=mrot)
            angle2 = kagan_angle(mt1, mt2)

            assert abs(angle - angle2) < eps

    def testRandomRotation(self):
        angles = []
        rstate = num.random.RandomState(10)
        for _ in range(100):
            mt1 = MomentTensor.random_mt(magnitude=-1.0)
            mt2 = mt1.random_rotated(10.0, rstate=rstate)
            angle = kagan_angle(mt1, mt2)
            angles.append(angle)

        assert abs(num.mean(angles) - 8.330) < 0.01

        mt1 = MomentTensor.random_mt(magnitude=-1.0)
        mt2 = mt1.random_rotated(angle=10.0)

        assert abs(kagan_angle(mt1, mt2) - 10.0) < 0.0001


if __name__ == "__main__":
    util.setup_logging('test_moment_tensor', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import range
import unittest
import numpy as num  # noqa
import logging
import shutil
from tempfile import mkdtemp
from pyrocko.dataset import crust2x2
from pyrocko import util

logger = logging.getLogger('pyrocko.test.test_crust2x2')


class Crust2x2TestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = mkdtemp('pyrocko.crust2x2')
        self.db = crust2x2.Crust2()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_crust2(self):
        nprofiles = 25
        lons = num.random.randint(-180, 180, nprofiles)
        lats = num.random.randint(-90, 90, nprofiles)
        for i in range(nprofiles):
            self.db.get_profile(lats[i], lons[i])

    def test_profiles(self):
        p = self.db.get_profile(25, 30)
        p.elevation()
        p.get_layer(1)
        p.get_weeded()
        p.averages()


if __name__ == "__main__":
    util.setup_logging('test_crust2x2', 'info')
    unittest.main()



from __future__ import division, print_function, absolute_import
import unittest
import numpy as num  # noqa
import logging
import shutil
from tempfile import mkdtemp
from os.path import join as pjoin

from pyrocko import util
from pyrocko.dataset import crustdb

logger = logging.getLogger('pyrocko.test.test_crustdb')


class CrustDBTestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = mkdtemp('pyrocko.crustdb')
        self.db = crustdb.CrustDB()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    @unittest.skip('Plotting not tested atm.')
    def test_map(self):
        tmpmap = pjoin(self.tmpdir, 'map.ps')
        self.db.plotMap(tmpmap)

    def test_selections(self):
        polygon = [(25., 30.), (30., 30.), (30, 25.), (25., 25.)]
        self.db.selectLocation(25., 95., 5.)
        self.db.selectRegion(25., 30., 20., 45.)
        self.db.selectPolygon(polygon)
        self.db.selectVs()
        self.db.selectVp()
        self.db.selectMaxDepth(40.)
        self.db.selectMinDepth(20.)

    @unittest.skip('Plotting not tested atm.')
    def test_ploting(self):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.gca()

        profile = self.db[10]
        self.db.plot(axes=ax)
        self.db.plotHistogram(axes=ax)
        profile.plot(axes=ax)

        fig.clear()


if __name__ == "__main__":
    util.setup_logging('test_crustdb', 'warning')
    unittest.main()
